import numpy as np
from scipy.optimize import fsolve


def _biexp(t, a, b, c, d):
    return a * (b * np.exp(-c * t) + (1 - b) * np.exp(-d * t))


def _biexp_root(t, a, b, c, d, val):
    """for fsolve"""
    return _biexp(t, a, b, c, d) - val


class Clearance_1m:
    """Class for the dose rate clearance function from a radioactive person."""

    def __init__(self, model, model_params, distance):
        """Supply an exponential or biexponential model of the
        dose rate at some distance (between 1 and 3 m) from the
        radioactive person as a function of the time from administration
        of the radioactive substance.

        The dose rate at 1 m is calculated assuming it
        drops off with distance as the inverse 1.5 power.

        Args:
            model (str): Type of clearance, either 'exponential' or 'biexponential'.
            model_params (list): Model parameters in terms of dose rate at ``distance``.
                If ``model`` is 'exponential', ``model_params`` is
                [initial dose rate (uSv/h) at ``distance``, effective half-life (h)].
                If ``model`` is 'biexponential', ``model_params`` is
                [initial dose rate (uSv/h) at ``distance``, fraction of first component (0 to 1),
                half-life (h) of first component, half-life (h) of second component].
            distance (int or float): Distance (m) from radioactive person.

        Raises:
            ValueError: If ``model`` not one of 'exponential' or 'biexponential'.
            TypeError: If ``model_params`` not list.
            ValueError: If ``model_params`` not length 2 for exponential.
            ValueError: If ``model_params`` not length 4 for biexponential.
            TypeError: If element in ``model_params`` not int or float.
            TypeError: If ``distance`` not int or float.
            ValueError: If ``distance`` not between 1 and 3 m, inclusive.
        """
        model_options = ["exponential", "biexponential"]
        if model not in model_options:
            raise ValueError("model not one of {}".format(model_options))
        if not isinstance(model_params, list):
            raise TypeError("model_params not list")
        if model == "exponential":
            if len(model_params) != 2:
                raise ValueError("model_params not length 2 for exponential")
        if model == "biexponential":
            if len(model_params) != 4:
                raise ValueError("model_params not length 4 for biexponential")
        if any(not isinstance(p, (int, float)) for p in model_params):
            raise TypeError("element in model_params not int or float")
        if not isinstance(distance, (int, float)):
            raise TypeError("distance not int or float")
        if not (1.0 <= distance <= 3.0):
            raise ValueError("distance not between 1 and 3 m, inclusive")

        self.model = model
        self.model_params = model_params.copy()
        self.model_params[0] = self.model_params[0] * (distance**1.5)

    def at_timedelta(self, timedelta, init=None):
        """Return the dose rate (uSv/h) at 1 m from the radioactive person at ``timedelta`` (h) after administration.
        If ``init`` is provided and is not None,
        instead evaluate at ``timedelta`` the clearance function with initial value ``init``.

        For example, set ``init`` to the administered activity to return the
        activity on board the person at ``timedelta``.

        Args:
            timedelta (int or float): Time (h) from administration.
            init (int or float, optional): Initial value of clearance function.
                Default is the initial dose rate at 1 m.

        Raises:
            TypeError: If ``timedelta`` not int or float.
            ValueError: If ``timedelta`` less than 0.
            TypeError: If ``init`` not None and not int or float.
            ValueError: If ``init`` not None and less than 0.

        Returns:
            float: The dose rate at 1 m from the radioactive person at ``timedelta``
            if ``init`` is None, else the clearance function
            with initial value ``init`` evaluated at ``timedelta``.
        """
        if not isinstance(timedelta, (int, float)):
            raise TypeError("timedelta not int or float")
        if timedelta < 0.0:
            raise ValueError("timedelta less than 0")

        params = self.model_params.copy()
        if init is not None:
            if not isinstance(init, (int, float)):
                raise TypeError("init not None and not int or float")
            if init < 0.0:
                raise ValueError("init not None and less than 0")
            params[0] = init

        if self.model == "exponential":
            return params[0] * (1 / 2) ** (timedelta / params[1])
        elif self.model == "biexponential":
            return _biexp(
                timedelta,
                params[0],
                params[1],
                np.log(2) / params[2],
                np.log(2) / params[3],
            )

    def get_timedelta(self, val, init=None):
        """Return the time (h) from administration to when the dose rate
        at 1 m from the radioactive person reaches ``val`` (uSv/h).
        If ``init`` is provided and is not None,
        the clearance function is used with initial value
        ``init`` instead of the initial dose rate at 1 m.

        For example, set ``init`` to the administered activity to return the time
        for the activity on board the radioactive person to reach ``val``.

        Args:
            val (int or float): Value to be reached by clearance function.
                If ``init`` is None, this is a dose rate (uSv/h) at 1 m.
            init (int or float, optional): Initial value of the clearance function.
                Default is the initial dose rate at 1 m.

        Raises:
            TypeError: If ``val`` not int or float.
            ValueError: If ``val`` not greater than 0.
            TypeError: If ``init`` not None and not int or float.
            ValueError: If ``init`` not None and not greater than 0.
            ValueError: If ``val`` greater than or equal to initial value
                of clearance function.

        Returns:
            float: Time (h) from administration to when the clearance
            function reaches ``val``.
        """
        if not isinstance(val, (int, float)):
            raise TypeError("val not int or float")
        if val <= 0.0:
            raise ValueError("val not greater than 0")

        params = self.model_params.copy()
        if init is not None:
            if not isinstance(init, (int, float)):
                raise TypeError("init not None and not int or float")
            if init <= 0.0:
                raise ValueError("init not None and not greater than 0")
            params[0] = init

        if val >= params[0]:
            raise ValueError(
                "val greater than or equal to initial value of clearance function"
            )

        # for biexp, try initial guesses 0 h, 24 h, 48 h, etc until fsolve finds a solution
        hrs0 = 0
        while True:
            if self.model == "exponential":
                hrs = params[1] * np.log(params[0] / val) / np.log(2)
            elif self.model == "biexponential":
                hrs = fsolve(
                    _biexp_root,
                    hrs0,
                    args=(
                        params[0],
                        params[1],
                        np.log(2) / params[2],
                        np.log(2) / params[3],
                        val,
                    ),
                )[0]

            if hrs != hrs0:
                break

            hrs0 += 24.0

        return hrs

    def residence_time(self):
        """Return the whole body residence time (a.k.a. time-integrated activity coefficient)
        for the radioactive person.

        Returns:
            float: Residence time (h).
        """
        if self.model == "exponential":
            return self.model_params[1] / np.log(2)
        elif self.model == "biexponential":
            fraction1 = self.model_params[1]
            return (
                (fraction1 * self.model_params[2])
                + ((1.0 - fraction1) * self.model_params[3])
            ) / np.log(2)



import pytest
import numpy as np
from datetime import datetime, timedelta
from glowgreen import ContactPatternOnceoff, Clearance_1m


def test_onceoff_init():
    theta = [0, 6]
    c = [1.5, 25]
    d = [0.1, 1]
    cpat = ContactPatternOnceoff(theta, c, d)
    assert np.array_equal(cpat.theta, np.array([0, 6]))
    assert np.allclose(cpat.c, np.array([1.5, 25]))
    assert isinstance(cpat.c, np.ndarray)
    assert np.array_equal(cpat.d, np.array([0.1, 1]))

    theta = 0
    c = 29
    d = 3
    cpat = ContactPatternOnceoff(theta, c, d)
    assert np.array_equal(cpat.theta, np.array([0]))
    assert np.array_equal(cpat.c, np.array([29]))
    assert np.array_equal(cpat.d, np.array([3]))

    theta = 1.6
    c = 29
    d = 3
    with pytest.raises(ValueError):
        cpat = ContactPatternOnceoff(theta, c, d)

    theta = 0
    c = 29
    d = 0
    with pytest.raises(ValueError):
        cpat = ContactPatternOnceoff(theta, c, d)


def test_get_dose_exp():
    """Test ContactPatternOnceoff.get_dose method with an exponential clearance function.

    Example problem:
      * Dose rate at 1 m is (50 uSv/h)*exp[-ln2 t / (40 h)], where t is time from admin.
      * Contact pattern is 6.5 hours at 0.3 m, 3 hour break, then 24 hours at 1.5 m.
    Caclulate the dose for a restriction period of 51 h.
    """
    dose_rate_init_1m = 50  # uSv/h
    effective_half_life = 40  # h
    tau = 51  # h

    # dose calculation by hand
    lmbda = np.log(2) / effective_half_life
    factor1 = dose_rate_init_1m / lmbda * np.exp(-lmbda * tau)
    factor2 = 1 / (0.3) ** (1.5) * np.exp(-lmbda * 0) * (1 - np.exp(-lmbda * 6.5))
    factor2 += 1 / (1.5) ** (1.5) * np.exp(-lmbda * 9.5) * (1 - np.exp(-lmbda * 24))
    dose_hand = factor1 * factor2 / 1000.0  # uSv -> mSv

    # dose calculation by code
    theta = np.array([0, 9.5])
    c = np.array([6.5, 24])
    d = [0.3, 1.5]
    cpat = ContactPatternOnceoff(theta, c, d)

    cfit = Clearance_1m("exponential", [dose_rate_init_1m, effective_half_life], 1.0)

    dose_code = cpat.get_dose(cfit, tau)
    assert np.isclose(dose_code, dose_hand)

    with pytest.raises(ValueError):
        cpat.get_dose(cfit, -0.1)

    assert cpat.get_dose(cfit, np.inf) == 0.0


def test_get_dose_biexp():
    """Test ContactPatternOnceoff.get_dose method with a biexponential clearance function.

    Example problem:
      * Dose rate at 1 m is (90 uSv/h)*[0.3*exp(-ln2 t / (5 h)) + 0.7*exp(-ln2 t / (50 h))],
        where t is time from admin.
      * Contact pattern is 6.5 hours at 0.3 m, 3 hour break, then 24 hours at 1.5 m.
    Caclulate the dose for a restriction period of 51 h.
    """
    dose_rate_init_1m = 90  # uSv/h
    fraction1 = 0.3
    half_life1 = 5  # h
    half_life2 = 50  # h
    tau = 51  # h

    # dose calculation by hand
    fraction2 = 1.0 - fraction1
    lmbda1 = np.log(2) / half_life1
    lmbda2 = np.log(2) / half_life2

    first_factor1 = dose_rate_init_1m * fraction1 / lmbda1 * np.exp(-lmbda1 * tau)
    first_factor2 = (
        1 / (0.3) ** (1.5) * np.exp(-lmbda1 * 0) * (1 - np.exp(-lmbda1 * 6.5))
    )
    first_factor2 += (
        1 / (1.5) ** (1.5) * np.exp(-lmbda1 * 9.5) * (1 - np.exp(-lmbda1 * 24))
    )

    second_factor1 = dose_rate_init_1m * fraction2 / lmbda2 * np.exp(-lmbda2 * tau)
    second_factor2 = (
        1 / (0.3) ** (1.5) * np.exp(-lmbda2 * 0) * (1 - np.exp(-lmbda2 * 6.5))
    )
    second_factor2 += (
        1 / (1.5) ** (1.5) * np.exp(-lmbda2 * 9.5) * (1 - np.exp(-lmbda2 * 24))
    )

    dose_hand = (
        (first_factor1 * first_factor2) + (second_factor1 * second_factor2)
    ) / 1000.0  # uSv -> mSv

    # dose calculation by code
    theta = np.array([0, 9.5])
    c = np.array([6.5, 24])
    d = [0.3, 1.5]
    cpat = ContactPatternOnceoff(theta, c, d)

    cfit = Clearance_1m(
        "biexponential", [dose_rate_init_1m, fraction1, half_life1, half_life2], 1.0
    )

    dose_code = cpat.get_dose(cfit, tau)
    assert np.isclose(dose_code, dose_hand)

    with pytest.raises(ValueError):
        cpat.get_dose(cfit, -0.1)

    assert cpat.get_dose(cfit, np.inf) == 0.0


def test_get_restriction_exp():
    """Test ContactPatternOnceoff.get_restriction and ContactPatternOnceoff._get_restriction_arrays methods for
    exponential clearance function.

    * ContactPatternOnceoff.get_restriction returns true restriction period
    * ContactPatternOnceoff._get_restriction_arrays returns ceil(true restriction period)

    Example problem:
      * Dose rate at 1 m is (50 uSv/h)*exp[-ln2 t / (40 h)], where t is time from admin.
      * Contact pattern is once-off; 6.5 hours at 0.3 m, 3 hour break, then 24 hours at 1.5 m.
    Calculate the dose for a delay of 51.73 h, then use this dose as a dose constraint
    and calculate the restriction period.
    """
    dose_rate_init_1m = 50  # uSv/h
    effective_half_life = 40  # h
    tau_original = 51.73  # h

    # dose calculation by hand
    lmbda = np.log(2) / effective_half_life
    factor1 = dose_rate_init_1m / lmbda * np.exp(-lmbda * tau_original)
    factor2 = 1 / (0.3) ** (1.5) * np.exp(-lmbda * 0) * (1 - np.exp(-lmbda * 6.5))
    factor2 += 1 / (1.5) ** (1.5) * np.exp(-lmbda * 9.5) * (1 - np.exp(-lmbda * 24))
    dose_hand = factor1 * factor2 / 1000.0  # uSv -> mSv
    dose_constraint = dose_hand

    # restriction period calculation by code
    theta = np.array([0, 9.5])
    c = np.array([6.5, 24])
    d = [0.3, 1.5]
    cpat = ContactPatternOnceoff(theta, c, d)

    datetime_admin = datetime(year=2021, month=10, day=25, hour=10, minute=15)

    cfit = Clearance_1m("exponential", [dose_rate_init_1m, effective_half_life], 1.0)

    tau, dose, tau_arr, dose_arr, datetime_end = cpat._get_restriction_arrays(
        cfit, dose_constraint, admin_datetime=datetime_admin
    )
    tau, dose, tau_arr, dose_arr, datetime_end = cpat._get_restriction_arrays(
        cfit, dose_constraint, datetime_admin
    )
    assert tau == np.ceil(tau_original)
    assert datetime_end == datetime_admin + timedelta(hours=tau)
    assert dose == cpat.get_dose(cfit, tau)
    assert dose <= dose_constraint
    assert tau_arr[-1] == tau
    assert dose_arr[-1] == dose
    assert len(tau_arr) == len(dose_arr)
    assert cpat._get_restriction_arrays(cfit, dose_constraint)[4] is None

    tau_fast, dose_fast, datetime_end_fast = cpat.get_restriction(
        cfit, dose_constraint, admin_datetime=datetime_admin
    )
    tau_fast, dose_fast, datetime_end_fast = cpat.get_restriction(
        cfit, dose_constraint, datetime_admin
    )
    assert np.isclose(tau_fast, tau_original)
    assert datetime_end_fast == datetime_admin + timedelta(hours=tau_fast)
    assert dose_fast == cpat.get_dose(cfit, tau_fast)
    assert np.isclose(dose_fast, dose_constraint)
    assert cpat.get_restriction(cfit, dose_constraint)[2] is None

    assert cpat._get_restriction_arrays(cfit, np.inf, datetime_admin)[0] == 0.0
    assert cpat._get_restriction_arrays(cfit, np.inf, datetime_admin)[
        1
    ] == cpat.get_dose(cfit, 0.0)
    assert (
        cpat._get_restriction_arrays(cfit, np.inf, datetime_admin)[4] == datetime_admin
    )
    assert cpat.get_restriction(cfit, np.inf, admin_datetime=datetime_admin)[0] == 0.0
    assert cpat.get_restriction(cfit, np.inf, admin_datetime=datetime_admin)[
        1
    ] == cpat.get_dose(cfit, 0.0)
    assert (
        cpat.get_restriction(cfit, np.inf, admin_datetime=datetime_admin)[2]
        == datetime_admin
    )

    with pytest.warns(UserWarning):
        val = cpat._exact_restriction_period(cfit, np.inf)
    assert val == 0.0

    with pytest.raises(ValueError):
        cpat._get_restriction_arrays(cfit, 0.0, datetime_admin)
    with pytest.raises(ValueError):
        cpat._get_restriction_arrays(cfit, -1.3, datetime_admin)
    with pytest.raises(ValueError):
        cpat.get_restriction(cfit, 0.0, admin_datetime=datetime_admin)
    with pytest.raises(ValueError):
        cpat.get_restriction(cfit, -1.3, admin_datetime=datetime_admin)
    with pytest.raises(ValueError):
        cpat._exact_restriction_period(cfit, 0.0)
    with pytest.raises(ValueError):
        cpat._exact_restriction_period(cfit, -1.3)


def test__get_restriction_arrays_biexp():
    """Test ContactPatternOnceoff.get_restriction and ContactPatternOnceoff._get_restriction_arrays methods for
    biexponential clearance function.

    * ContactPatternOnceoff.get_restriction returns true restriction period
    * ContactPatternOnceoff._get_restriction_arrays returns ceil(true restriction period)

    Example problem:
      * Dose rate at 1 m is (90 uSv/h)*[0.3*exp(-ln2 t / (5 h)) + 0.7*exp(-ln2 t / (50 h))],
        where t is time from admin.
      * Contact pattern is once-off; 6.5 hours at 0.3 m, 3 hour break, then 24 hours at 1.5 m.
    Calculate the dose for a delay of 50.32 h, then use this dose as a dose constraint
    and calculate the restriction period.
    """
    dose_rate_init_1m = 90  # uSv/h
    fraction1 = 0.3
    half_life1 = 5  # h
    half_life2 = 50  # h
    tau_original = 50.32  # h

    # dose calculation by hand
    fraction2 = 1.0 - fraction1
    lmbda1 = np.log(2) / half_life1
    lmbda2 = np.log(2) / half_life2

    first_factor1 = (
        dose_rate_init_1m * fraction1 / lmbda1 * np.exp(-lmbda1 * tau_original)
    )
    first_factor2 = (
        1 / (0.3) ** (1.5) * np.exp(-lmbda1 * 0) * (1 - np.exp(-lmbda1 * 6.5))
    )
    first_factor2 += (
        1 / (1.5) ** (1.5) * np.exp(-lmbda1 * 9.5) * (1 - np.exp(-lmbda1 * 24))
    )

    second_factor1 = (
        dose_rate_init_1m * fraction2 / lmbda2 * np.exp(-lmbda2 * tau_original)
    )
    second_factor2 = (
        1 / (0.3) ** (1.5) * np.exp(-lmbda2 * 0) * (1 - np.exp(-lmbda2 * 6.5))
    )
    second_factor2 += (
        1 / (1.5) ** (1.5) * np.exp(-lmbda2 * 9.5) * (1 - np.exp(-lmbda2 * 24))
    )

    dose_hand = (
        (first_factor1 * first_factor2) + (second_factor1 * second_factor2)
    ) / 1000.0  # uSv -> mSv
    dose_constraint = dose_hand

    # restriction period calculation by code
    theta = np.array([0, 9.5])
    c = np.array([6.5, 24])
    d = [0.3, 1.5]
    cpat = ContactPatternOnceoff(theta, c, d)

    datetime_admin = datetime(year=2021, month=10, day=25, hour=10, minute=15)

    cfit = Clearance_1m(
        "biexponential", [dose_rate_init_1m, fraction1, half_life1, half_life2], 1.0
    )

    tau, dose, tau_arr, dose_arr, datetime_end = cpat._get_restriction_arrays(
        cfit, dose_constraint, admin_datetime=datetime_admin
    )
    tau, dose, tau_arr, dose_arr, datetime_end = cpat._get_restriction_arrays(
        cfit, dose_constraint, datetime_admin
    )
    assert tau == np.ceil(tau_original)
    assert datetime_end == datetime_admin + timedelta(hours=tau)
    assert dose == cpat.get_dose(cfit, tau)
    assert dose <= dose_constraint
    assert tau_arr[-1] == tau
    assert dose_arr[-1] == dose
    assert len(tau_arr) == len(dose_arr)
    assert cpat._get_restriction_arrays(cfit, dose_constraint)[4] is None

    tau_fast, dose_fast, datetime_end_fast = cpat.get_restriction(
        cfit, dose_constraint, admin_datetime=datetime_admin
    )
    assert np.isclose(tau_fast, tau_original)
    assert datetime_end_fast == datetime_admin + timedelta(hours=tau_fast)
    assert dose_fast == cpat.get_dose(cfit, tau_fast)
    assert np.isclose(dose_fast, dose_constraint)
    assert cpat.get_restriction(cfit, dose_constraint)[2] is None

    assert cpat._get_restriction_arrays(cfit, np.inf, datetime_admin)[0] == 0.0
    assert cpat._get_restriction_arrays(cfit, np.inf, datetime_admin)[
        1
    ] == cpat.get_dose(cfit, 0.0)
    assert (
        cpat._get_restriction_arrays(cfit, np.inf, datetime_admin)[4] == datetime_admin
    )
    assert cpat.get_restriction(cfit, np.inf, admin_datetime=datetime_admin)[0] == 0.0
    assert cpat.get_restriction(cfit, np.inf, admin_datetime=datetime_admin)[
        1
    ] == cpat.get_dose(cfit, 0.0)
    assert (
        cpat.get_restriction(cfit, np.inf, admin_datetime=datetime_admin)[2]
        == datetime_admin
    )

    with pytest.raises(ValueError):
        cpat._get_restriction_arrays(cfit, 0.0, datetime_admin)
    with pytest.raises(ValueError):
        cpat._get_restriction_arrays(cfit, -1.3, datetime_admin)
    with pytest.raises(ValueError):
        cpat.get_restriction(cfit, 0.0, admin_datetime=datetime_admin)
    with pytest.raises(ValueError):
        cpat.get_restriction(cfit, -1.3, admin_datetime=datetime_admin)



# Configuration file for the Sphinx documentation builder.
#
# For the full list of built-in configuration values, see the documentation:
# https://www.sphinx-doc.org/en/master/usage/configuration.html

# -- Project information -----------------------------------------------------
# https://www.sphinx-doc.org/en/master/usage/configuration.html#project-information


from glowgreen import __version__ as glowgreen_version

import os
import sys
sys.path.insert(0, os.path.abspath('../..'))
sys.path.insert(0, os.path.abspath('../../src/'))
sys.path.insert(0, os.path.abspath('../../src/glowgreen/'))

project = 'glowgreen'
copyright = '2022, South Australia Medical Imaging'
author = 'Jake Forster'
release = glowgreen_version

# -- General configuration ---------------------------------------------------
# https://www.sphinx-doc.org/en/master/usage/configuration.html#general-configuration

extensions = ['sphinx.ext.autodoc', 'sphinx.ext.napoleon']

autoclass_content = 'both'

templates_path = ['_templates']
exclude_patterns = []



# -- Options for HTML output -------------------------------------------------
# https://www.sphinx-doc.org/en/master/usage/configuration.html#options-for-html-output

html_theme = 'classic'
html_static_path = ['_static']



from datetime import datetime
from glowgreen import Clearance_1m, cs_restrictions, cs_patterns, restrictions_for


def test_cs_patterns():
    cs_patterns()


def test_restrictions_for():
    df_base = cs_patterns()

    admin_datetime = datetime(year=2021, month=10, day=25, hour=10, minute=15)

    model = "exponential"
    dose_rate_xm_init = 60.0
    effective_half_life = 11.0
    model_parameters = [dose_rate_xm_init, effective_half_life]
    measurement_distance = 2.0
    cfit = Clearance_1m(model, model_parameters, measurement_distance)

    num_treatments_in_year = 4

    df = restrictions_for(
        df_base, cfit, num_treatments_in_year, admin_datetime=admin_datetime
    )

    required_cols = ["restriction_period", "dose", "datetime_end"]
    for col in required_cols:
        assert col in df.columns


def test_cs_restrictions():
    admin_datetime = datetime(year=2021, month=10, day=25, hour=10, minute=15)

    model = "exponential"
    dose_rate_xm_init = 60.0
    effective_half_life = 11.0
    model_parameters = [dose_rate_xm_init, effective_half_life]
    measurement_distance = 2.0
    cfit = Clearance_1m(model, model_parameters, measurement_distance)

    num_treatments_in_year = 4

    cs_restrictions(cfit, num_treatments_in_year, admin_datetime)



import numpy as np
from glowgreen import Clearance_1m


def test_at_timedelta_exp():
    dose_rate_xm_init = 80.0
    effective_half_life = 11.0
    measurement_distance = 2.0
    n_half_lives = 1.3

    dose_rate_1m_init = dose_rate_xm_init * measurement_distance**1.5
    val_hand = dose_rate_1m_init * (1 / 2) ** (n_half_lives)

    model = "exponential"
    model_params = [dose_rate_xm_init, effective_half_life]
    cfit = Clearance_1m(model, model_params, measurement_distance)
    hrs = n_half_lives * effective_half_life
    val_code = cfit.at_timedelta(hrs)

    assert val_code == val_hand


def test_at_timedelta_biexp():
    dose_rate_xm_init = 50.0
    fraction_1 = 0.4
    half_life_1 = 11.0
    half_life_2 = 20.0
    measurement_distance = 2.0
    hrs = 42.1

    dose_rate_1m_init = dose_rate_xm_init * measurement_distance**1.5
    val_hand = dose_rate_1m_init * (
        (fraction_1 * (1 / 2) ** (hrs / half_life_1))
        + ((1 - fraction_1) * (1 / 2) ** (hrs / half_life_2))
    )

    model = "biexponential"
    model_params = [dose_rate_xm_init, fraction_1, half_life_1, half_life_2]
    cfit = Clearance_1m(model, model_params, measurement_distance)
    val_code = cfit.at_timedelta(hrs)

    assert np.isclose(val_code, val_hand)


def test_at_timedelta_activity_exp():
    a0 = 3e3
    n_half_lives = 1.3

    val_hand = a0 * (1 / 2) ** (n_half_lives)

    dose_rate_xm_init = np.nan
    effective_half_life = 11.0
    measurement_distance = 3.0
    model = "exponential"
    model_params = [dose_rate_xm_init, effective_half_life]
    measurement_distance = 3.0
    cfit = Clearance_1m(model, model_params, measurement_distance)
    hrs = n_half_lives * effective_half_life
    val_code = cfit.at_timedelta(hrs, init=a0)

    assert val_code == val_hand


def test_at_timedelta_activity_biexp():
    a0 = 3e3
    fraction_1 = 0.4
    half_life_1 = 11.0
    half_life_2 = 20.0
    hrs = 42.1

    val_hand = a0 * (
        (fraction_1 * (1 / 2) ** (hrs / half_life_1))
        + ((1 - fraction_1) * (1 / 2) ** (hrs / half_life_2))
    )

    dose_rate_xm_init = np.nan
    model = "biexponential"
    model_params = [dose_rate_xm_init, fraction_1, half_life_1, half_life_2]
    measurement_distance = 3.0
    cfit = Clearance_1m(model, model_params, measurement_distance)
    val_code = cfit.at_timedelta(hrs, init=a0)

    assert np.isclose(val_code, val_hand)


def test_get_timedelta_exp():
    dose_rate_xm_init = 80.0
    effective_half_life = 11.0
    measurement_distance = 2.0

    dose_rate_1m_init = dose_rate_xm_init * measurement_distance**1.5
    hrs_hand = effective_half_life * np.log(dose_rate_1m_init / 25.0) / np.log(2)

    model = "exponential"
    model_params = [dose_rate_xm_init, effective_half_life]
    cfit = Clearance_1m(model, model_params, measurement_distance)
    hrs_code = cfit.get_timedelta(25.0)

    assert hrs_code == hrs_hand


def test_get_timedelta_biexp():
    # fake biexp
    dose_rate_xm_init = 50.0
    fraction_1 = 1.0
    half_life_1 = 11.0
    half_life_2 = 20.0
    measurement_distance = 2.0

    dose_rate_1m_init = dose_rate_xm_init * measurement_distance**1.5
    hrs_hand = half_life_1 * np.log(dose_rate_1m_init / 25.0) / np.log(2)

    model = "biexponential"
    model_params = [dose_rate_xm_init, fraction_1, half_life_1, half_life_2]
    cfit = Clearance_1m(model, model_params, measurement_distance)
    hrs_code = cfit.get_timedelta(25.0)

    assert np.isclose(hrs_code, hrs_hand)

    model_params = [dose_rate_xm_init, 0.0, half_life_2, half_life_1]
    cfit = Clearance_1m(model, model_params, measurement_distance)
    hrs_code = cfit.get_timedelta(25.0)

    assert np.isclose(hrs_code, hrs_hand)


def test_get_timedelta_activity_exp():
    a0 = 3e3
    activity_limit_for_discharge = 5e2
    effective_half_life = 11.0

    hrs_hand = (
        effective_half_life * np.log(a0 / activity_limit_for_discharge) / np.log(2)
    )

    dose_rate_xm_init = np.nan
    measurement_distance = 3.0
    model = "exponential"
    model_params = [dose_rate_xm_init, effective_half_life]
    cfit = Clearance_1m(model, model_params, measurement_distance)
    hrs_code = cfit.get_timedelta(activity_limit_for_discharge, init=a0)

    assert hrs_code == hrs_hand


def test_get_timedelta_activity_biexp():
    # fake biexp
    a0 = 3e3
    activity_limit_for_discharge = 5e2
    fraction_1 = 1.0
    half_life_1 = 11.0
    half_life_2 = 20.0

    hrs_hand = half_life_1 * np.log(a0 / activity_limit_for_discharge) / np.log(2)

    dose_rate_xm_init = np.nan
    measurement_distance = 3.0
    model = "biexponential"
    model_params = [dose_rate_xm_init, fraction_1, half_life_1, half_life_2]
    cfit = Clearance_1m(model, model_params, measurement_distance)
    hrs_code = cfit.get_timedelta(activity_limit_for_discharge, init=a0)

    model_params = [dose_rate_xm_init, 0.0, half_life_2, half_life_1]
    cfit = Clearance_1m(model, model_params, measurement_distance)
    hrs_code = cfit.get_timedelta(activity_limit_for_discharge, init=a0)

    assert np.isclose(hrs_code, hrs_hand)


def test_residence_time_exp():
    model = "exponential"
    dose_rate_xm_init = 80.0
    effective_half_life = 11.0
    model_params = [dose_rate_xm_init, effective_half_life]
    measurement_distance = 2.0
    cfit = Clearance_1m(model, model_params, measurement_distance)

    assert cfit.residence_time() == effective_half_life / np.log(2)


def test_residence_time_biexp1():
    # fake biexp
    model = "biexponential"
    dose_rate_xm_init = 80.0
    fraction_1 = 1.0
    half_life_1 = 11.0
    half_life_2 = 20.0
    model_params = [dose_rate_xm_init, fraction_1, half_life_1, half_life_2]
    measurement_distance = 2.0
    cfit = Clearance_1m(model, model_params, measurement_distance)

    assert cfit.residence_time() == half_life_1 / np.log(2)


def test_residence_time_biexp2():
    model = "biexponential"
    dose_rate_xm_init = 80.0
    fraction_1 = 0.3
    half_life_1 = 11.0
    half_life_2 = 30.0
    model_params = [dose_rate_xm_init, fraction_1, half_life_1, half_life_2]
    measurement_distance = 2.0
    cfit = Clearance_1m(model, model_params, measurement_distance)

    assert cfit.residence_time() == (
        fraction_1 * half_life_1 + (1.0 - fraction_1) * half_life_2
    ) / np.log(2)



from datetime import datetime, date, time, timedelta
import numpy as np
import warnings
from . import clearance
import matplotlib.pyplot as plt
import pandas as pd
from scipy.optimize import fsolve


class _ContactPattern:
    """Base class for :class:`ContactPatternRepeating` and
    :class:`ContactPatternOnceoff`.
    Not for instantiation.

    Contains methods common to both repeating and onceoff contact patterns.
    """

    def __init__(self, theta, c, d):
        """Subclass constructors first `super` this constructor to perform
        non-specific quality control.

        Args:
            theta (int, float, list or numpy.ndarray): Start times (h) of pattern elements.
            c (int, float, list or numpy.ndarray): Durations (h) of pattern elements.
            d (int, float, list or numpy.ndarray): Distances (m) of pattern elements.

        Raises:
            TypeError: If ``theta`` is list or numpy.ndarray and ``c`` or `d` are not.
            ValueError: If ``theta``, ``c`` and ``d`` are not equal length.
            TypeError: If ``theta``, ``c`` or ``d`` contain a value not int, float, numpy.int32, numpy.int64 or numpy.float64.
            ValueError: If negative value in ``theta``, ``c`` or ``d``.
            ValueError: If pattern elements not provided in order.
            ValueError: If pattern elements overlap.
            TypeError: If ``theta`` is int or float and ``c`` or ``d`` are not.
            ValueError: If ``theta``, ``c`` or ``d`` is negative.
            TypeError: If ``theta`` not type int, float, list or numpy.ndarray.
        """
        if isinstance(theta, (list, np.ndarray)):
            if (not isinstance(c, (list, np.ndarray))) or (
                not isinstance(d, (list, np.ndarray))
            ):
                raise TypeError("theta is list or numpy.ndarray and c or d are not")

            if (len(theta) != len(c)) or (len(theta) != len(d)):
                raise ValueError("theta, c and d not equal length")

            if not all(
                all(
                    isinstance(x, (int, float, np.int32, np.int64, np.float64))
                    for x in a
                )
                for a in [theta, c, d]
            ):
                raise TypeError(
                    "theta, c or d contain a value not int, float, numpy.int32, numpy.int64 or numpy.float64"
                )

            if any(any(x < 0.0 for x in a) for a in [theta, c, d]):
                raise ValueError("negative value in theta, c or d")

            if any(theta[i + 1] < theta[i] for i in range(len(theta) - 1)):
                raise ValueError("pattern elements not provided in order")

            for i in range(len(theta) - 1):
                if (theta[i] + c[i]) > theta[i + 1]:
                    raise ValueError("pattern elements overlap")

            self.theta = theta.copy()
            self.c = c.copy()
            self.d = d.copy()
        elif isinstance(theta, (int, float)):
            if (not isinstance(c, (int, float))) or (not isinstance(d, (int, float))):
                raise TypeError("theta is int or float and c or d are not")

            if any(x < 0.0 for x in [theta, c, d]):
                raise ValueError("theta, c or d is negative")

            self.theta = theta
            self.c = c
            self.d = d
        else:
            raise TypeError("theta not type int, float, list or numpy.ndarray")

    @staticmethod
    def _dist_func(d):
        """Return the factor by which the dose rate at 1 m can be multiplied to get the dose rate
        at some other distance during the pattern.

        Inverse 1.5 power is used.

        Args:
            d (int, float or numpy.ndarray): Distance (m) of pattern element.

        Returns:
            float or numpy.ndarray: Correction factor to go from dose rate at 1 m to dose rate at ``d``.
        """
        return (1 / d) ** 1.5

    def plot(
        self,
        name=None,
        cfit: clearance.Clearance_1m = None,
        dose_constraint=None,
        admin_datetime: datetime = None,
        test=False,
    ):
        """Generate a plot of the contact pattern and, if sufficient information is provided in optional args,
        an additional plot of the dose from the pattern as a function of the
        time from administration to beginning sharing the onceoff pattern, or
        resuming sharing the repeating pattern, with a radioactive person.

        Args:
            name (str, optional): A name for the contact pattern, to appear in the figure title.
            cfit (clearance.Clearance_1m, optional): Dose rate clearance from radioactive person.
            dose_constraint (int or float, optional): Dose constraint (mSv).
            admin_datetime (datetime.datetime, optional): Administration datetime.
                Has no effect if pattern is onceoff.
            test (bool, optional): Option to not show the plot. Used for testing. Default is False.
        """
        if None not in [cfit, dose_constraint] and (
            admin_datetime is not None or isinstance(self, ContactPatternOnceoff)
        ):
            fig, (ax1, ax2) = plt.subplots(2, 1)
            if name is not None:
                fig.suptitle(name, fontsize=11)
        else:
            fig, ax1 = plt.subplots()
            if name is not None:
                ax1.set_title(name, fontsize=11)

        if isinstance(self, ContactPatternRepeating):
            ax1.bar(
                self.theta,
                self.d ** (-3 / 2),
                width=self.c,
                align="edge",
                fill=True,
                edgecolor="black",
                color=(0, 158 / 255, 115 / 255),
            )
            ax1.set_xlabel("24-hour time")
            ax1.set_xlim(left=0.0, right=24.0)
            ax1.set_xticks(np.arange(0, 25, 1))
            y_up = ax1.get_ylim()[1]
            if admin_datetime is not None:
                ax1.annotate(
                    "ADMIN",
                    xy=(admin_datetime.hour + admin_datetime.minute / 60.0, 0),
                    xytext=(
                        admin_datetime.hour + admin_datetime.minute / 60.0,
                        0.2 * y_up,
                    ),
                    horizontalalignment="center",
                    arrowprops={"arrowstyle": "->"},
                )
        elif isinstance(self, ContactPatternOnceoff):
            ax1.bar(
                self.theta,
                self.d ** (-3 / 2),
                width=self.c,
                align="edge",
                edgecolor="black",
                color=(86 / 255, 180 / 255, 233 / 255),
                fill=True,
            )
            ax1.set_xlabel("Time from end of restriction (h)")
            ax1.set_xlim(left=0.0)
        ax1.set_ylabel(r"${[(1~\mathrm{m}) ~/~ \mathrm{distance}]}^{1.5}$")
        ax1.set_ylim(bottom=1e-4)

        warnings.filterwarnings("ignore", "divide by zero encountered in power")
        secax = ax1.secondary_yaxis(
            "right", functions=(lambda x: x ** (-2 / 3), lambda x: x ** (-3 / 2))
        )

        secax.set_yticks(np.unique(self.d))
        secax.set_ylabel("Distance (m)")

        if None not in [cfit, dose_constraint] and (
            admin_datetime is not None or isinstance(self, ContactPatternOnceoff)
        ):
            _, _, tau_arr, dose_arr, _ = self._get_restriction_arrays(
                cfit, dose_constraint, admin_datetime
            )
            ax2.plot(
                tau_arr,
                dose_arr,
                "o",
                markersize=4,
                markerfacecolor="None",
                markeredgecolor="black",
            )
            ax2.plot(
                tau_arr,
                dose_constraint * np.ones(len(tau_arr)),
                color=(204 / 255, 121 / 255, 167 / 255),
                ls="--",
                label="{:g} mSv".format(dose_constraint),
            )
            ax2.set_xlabel("Delay from time of administration (h)")
            ax2.set_ylabel("Dose (mSv)")
            ax2.set_xlim(left=0.0)
            ax2.set_ylim(bottom=0.0)
            ax2.legend(loc="best")
            plt.tight_layout()

        if not test:
            plt.show()


class ContactPatternRepeating(_ContactPattern):
    """Class for infinitely repeating diurnal contact patterns."""

    def __init__(self, theta, c, d):
        """Constructor:
         * Calls constructor of :class:`_ContactPattern`.
         * ``theta``, ``c`` and ``d`` are converted into numpy.ndarrays if they were not supplied as such.
         * Pattern elements with duration longer than 1 h are broken up, so that
           the restriction period can be better resolved.

        Args:
            theta (int, float, list or numpy.ndarray): Time (h) from 12 AM to start of pattern element.
            c (int, float, list or numpy.ndarray): Duration (h) of pattern element.
            d (int, float, list or numpy.ndarray): Distance (m) of pattern element;
                i.e., distance from radioactive person for duration of pattern element.

        Raises:
            ValueError: If element of repeating pattern has ``d`` of 0 and ``c`` not 0.
            ValueError: If repeating pattern extends beyond pattern period (24 h).
        """
        super().__init__(theta, c, d)

        self.t_r = time(
            0, 0
        )  # Reference time (12 AM), which theta is defined with respect to.
        self.p = 24.0  # Pattern period (h)

        if isinstance(self.theta, (list, np.ndarray)):
            if any(c_num != 0 and d_num == 0 for c_num, d_num in zip(self.c, self.d)):
                raise ValueError("element of repeating pattern has d of 0 and c not 0")
            if (self.theta[-1] + self.c[-1]) > self.p:
                raise ValueError(
                    "repeating pattern extends beyond pattern period (24 h)"
                )
        else:
            if self.c != 0 and self.d == 0:
                raise ValueError("element of repeating pattern has d of 0 and c not 0")
            if (self.theta + self.c) > self.p:
                raise ValueError(
                    "repeating pattern extends beyond pattern period (24 h)"
                )

        self._graduate_pattern()

    @staticmethod
    def _graduate_pattern_element(
        theta_num, c_num, d_num
    ) -> tuple[np.ndarray, np.ndarray, np.ndarray]:
        """Writes a single pattern element as 1 or more elements,
        with all but the last element having duration 1 h.

        For example, ``c_num`` = 1.2 -> c_subarr = np.array([1., 0.2]),
        theta_subarr = np.array([``theta_num``, ``theta_num`` + 1.]) and
        d_subarr = np.array([``d_num``, ``d_num``]).

        Args:
            theta_num (int or float): Time (h) from 12 AM to start of pattern element.
            c_num (int or float): Duration (h) of pattern element.
            d_num (int or float): Distance (m) of pattern element.

        Returns:
            tuple[numy.ndarray, numy.ndarray, numy.ndarray]:
            Graduated versions of ``theta_num``, ``c_num`` and ``d_num``.
        """
        theta_subarr = np.linspace(
            theta_num,
            theta_num + np.ceil(c_num),
            num=int(np.ceil(c_num)),
            endpoint=False,
        )
        c_subarr = np.ones(int(np.floor(c_num)))
        if len(c_subarr) < len(theta_subarr):
            c_subarr = np.append(c_subarr, c_num - np.floor(c_num))
        d_subarr = d_num * np.ones(len(theta_subarr))
        return theta_subarr, c_subarr, d_subarr

    def _graduate_pattern(self):
        """Pattern elements with duration longer than 1 h are broken up.

        ``self``.theta, ``self``.c and ``self``.d finish as numpy.ndarrays.
        """
        if isinstance(self.theta, (int, float)):
            self.theta, self.c, self.d = self._graduate_pattern_element(
                self.theta, self.c, self.d
            )
        else:
            theta_subarr_list = []
            c_subarr_list = []
            d_subarr_list = []
            for i in range(len(self.theta)):
                theta_subarr, c_subarr, d_subarr = self._graduate_pattern_element(
                    self.theta[i], self.c[i], self.d[i]
                )
                theta_subarr_list.append(theta_subarr)
                c_subarr_list.append(c_subarr)
                d_subarr_list.append(d_subarr)
            self.theta = np.array(
                [item for subarr in theta_subarr_list for item in subarr]
            )
            self.c = np.array([item for subarr in c_subarr_list for item in subarr])
            self.d = np.array([item for subarr in d_subarr_list for item in subarr])

    def _subtract_next_t_r_from(self, admin_datetime):
        """Return ``admin_datetime`` minus the next reference time (h).

        Args:
            admin_datetime (datetime.datetime): Administration datetime.

        Returns:
            float: ``admin_datetime`` minus the next reference time (h). Range is [-``self``.p, 0[.
        """
        dt_next_t_r = datetime.combine(
            date(
                day=admin_datetime.day,
                month=admin_datetime.month,
                year=admin_datetime.year,
            ),
            self.t_r,
        ) + timedelta(hours=self.p)
        hrs = (admin_datetime - dt_next_t_r).total_seconds() / 3600.0
        return hrs

    def _get_phi(self, admin_datetime, tau):
        """Return the shift (h) from reference time (12 AM) to when the pattern is resumed (i.e. when ``tau`` ends)

        Args:
            admin_datetime (datetime.datetime): Administration datetime.
            tau (int or float): The time (h) from ``admin_datetime`` to when the pattern is resumed.

        Returns:
            float: The shift (h) from reference time (12 AM) to when the pattern is resumed.
        """
        dt = self._subtract_next_t_r_from(admin_datetime)
        return (dt % (tau + self.p)) % self.p

    def _if_neg_add_p(self, x):
        y = np.zeros(len(x))
        for i in range(len(x)):
            if x[i] >= 0:
                y[i] = x[i]
            else:
                y[i] = x[i] + self.p
            # When the delay period is increased so that its end coincides with the start of a pattern element,
            # x is sometimes a tiny negative number, so p is returned when 0 should be returned
            if np.isclose(y[i], self.p) or np.isclose(y[i], 0.0):
                y[i] = 0.0
        return y

    def _time_to_end_of_element(self, tau, admin_datetime: datetime):
        """If the delay ``tau`` ends during a pattern element, return the time (h) til the end of the element,
        else return 0.

        Args:
            tau (int or float): The time (h) from ``admin_datetime`` to when the pattern is resumed.
            admin_datetime (datetime.datetime): Administration datetime.

        Returns:
            float: The time (h) from when ``tau`` ends til end of concurrent pattern element. Range is [0, 1[.
        """
        t_left = 0.0
        phi = self._get_phi(admin_datetime, tau)
        for i in range(len(self.theta)):
            if self.theta[i] < phi < (self.theta[i] + self.c[i]):
                t_left = self.theta[i] + self.c[i] - phi
                break
        return t_left

    def get_dose(
        self, cfit: clearance.Clearance_1m, tau, admin_datetime: datetime
    ) -> tuple[float, float]:
        """Calculate the lifetime dose (mSv) from sharing the infinitely repeating pattern
        with a radioactive person, for a given dose rate clearance function,
        delay period and administration datetime.

        Implements method in: Cormack J & Shearer J. "Calculation of radiation exposures from patients to whom
        radioactive materials have been administered." Phys Med Biol 1998; 43(3).

        NB. If the delay period ``tau`` ends during a pattern element,
        the dose calculated will not include a contribution from the
        remainder of that pattern element in the first cycle.
        I.e., the calculated dose will be for the delay period shifted forward (by less than 1 h)
        to the end of the pattern element.
        Hence, the potentially corrected delay period is returned along with the dose.

        Args:
            cfit (clearance.Clearance_1m): Dose rate clearance from radioactive person.
            tau (int or float): Delay period (h); i.e., the time
                from ``admin_datetime`` to when the pattern is resumed.
            admin_datetime (datetime.datetime): Administration datetime.

        Raises:
            ValueError: If ``tau`` less than 0.
            ValueError: If ``tau`` is numpy.inf with repeating pattern.

        Returns:
            tuple[float, float]: Dose (mSv) from the pattern from the end of the delay period to infinity,
            and the potentially corrected delay period (h).
        """
        if tau < 0.0:
            raise ValueError("tau less than 0")
        if tau == np.inf:  # cannot handle, e.g. _get_phi
            raise ValueError("tau is numpy.inf with repeating pattern")

        if cfit.model == "exponential":
            dose_rate_1m_init, effective_half_life = cfit.model_params
            a = 1
            lmbda = np.log(2) / effective_half_life
            lmbda = np.array(
                [lmbda]
            )  # make it an ndarray of length n for n-component exponential
        elif cfit.model == "biexponential":
            dose_rate_1m_init, fraction_1, half_life_1, half_life_2 = cfit.model_params
            fraction_2 = 1.0 - fraction_1
            lmbda_1 = np.log(2) / half_life_1
            lmbda_2 = np.log(2) / half_life_2
            a = np.array([fraction_1, fraction_2])
            lmbda = np.array([lmbda_1, lmbda_2])

        phi = self._get_phi(admin_datetime, tau)
        theta_prime = self._if_neg_add_p(
            self.theta - phi
        )  # theta if the reference time was the end of the delay

        arr = np.zeros(len(lmbda))
        for i in range(len(lmbda)):
            # summing over theta_prime, self.c, self.d arrays (if more than 1 element in pattern, else np.sum does nothing)
            arr[i] = np.sum(
                self._dist_func(self.d)
                * np.exp(-lmbda[i] * theta_prime)
                * (1 - np.exp(-lmbda[i] * self.c))
            )

        # summing over a, lmbda, arr arrays (if biexp, otherwise np.sum does nothing)
        dose = dose_rate_1m_init * np.sum(
            (a / lmbda)
            * np.exp(-lmbda * tau)
            * (1 / (1 - np.exp(-lmbda * self.p)))
            * arr
        )

        tau += self._time_to_end_of_element(tau, admin_datetime)

        return dose / 1000.0, tau  # uSv -> mSv

    def get_dose_finite(
        self, cfit: clearance.Clearance_1m, t1, t2, admin_datetime: datetime
    ) -> tuple[float, float, float]:
        """Return the dose (mSv) from sharing the repeating pattern with a radioactive person between 2 time points.

        For example, the dose if the pattern is resumed after some delay period
        then ceased again permanently at a later time.

        NB. If ``t1`` or ``t2`` end during a pattern element, the calculated dose is for
        that time point shifted forward (by less than 1 h) to the end of the pattern element.
        Hence, the potentially corrected time interval is returned along with the dose.

        Args:
            cfit (clearance.Clearance_1m): Dose rate clearance from radiaoctive person.
            t1 (int or float): The time (h) from ``admin_datetime`` to the start of the exposure.
            t2 (int or float): The time (h) from ``admin_datetime`` to the end of the exposure.
            admin_datetime (datetime.datetime): Administration datetime.

        Raises:
            ValueError: If ``t2`` less than ``t1``.

        Returns:
            tuple[float, float, float]: Dose (mSv) from the pattern between 2 time points
            with respect to ``admin_datetime``,
            first time point (h), and second time point (h).
        """
        if t2 < t1:
            raise ValueError("t2 less than t1")
        d1, t1 = self.get_dose(cfit, t1, admin_datetime)
        d2, t2 = self.get_dose(cfit, t2, admin_datetime)
        dose_finite = d1 - d2
        return dose_finite, t1, t2

    def get_restriction(
        self,
        cfit: clearance.Clearance_1m,
        dose_constraint,
        admin_datetime: datetime,
        next_element=True,
    ) -> tuple[float, float, datetime]:
        """Calculate the restriction period; i.e., the least time from administration
        (up to the resolution of the pattern element widths)
        to when sharing the pattern with a radioactive person can be resumed, such that the lifetime dose from the
        infinitely repeating pattern is less than the dose constraint.

        NB. Using ``next_element`` True is advisable and can make a large difference for sparse repeating patterns such as public transport to and from work.
        For example, if we want you to miss the morning bus, it means you have to wait for the afternoon bus;
        you can't just get on the morning bus at 9 AM instead of 8 AM and expect to comply with the dose constraint.
        By always having the end of the restriction period coincide with the start of a pattern element,
        it is clear that contact can resume immediately at the end of the restriction period.

        Args:
            cfit (clearance.Clearance_1m): Dose rate clearance from radioactive person.
            dose_constraint (int or float): Dose constraint (mSv).
            admin_datetime (datetime.datetime): Administration datetime.
            next_element (bool, optional): If the end of the restriction period does not coincide with the start of a pattern element,
                extend the restriction period to the start of the next pattern element. Default is True.

        Raises:
            ValueError: If ``dose_constraint`` not greater than 0.

        Returns:
            tuple[float, float, datetime.datetime]: Calculated restriction period (h),
            dose (mSv) from the pattern from the end of this restriction period to infinity,
            and the datetime at the end of this restriction period.

            With ``next_element`` False, the calculated restriction period is in the range:
            [true restriction period, true restriction period + (1 h) - min(``self``.c)[,
            where the true restriction period is limited only by the pattern element widths.
        """
        if dose_constraint <= 0.0:
            raise ValueError("dose_constraint not greater than 0")

        tau = 0.0
        dose, tau = self.get_dose(cfit, tau, admin_datetime)
        while dose > dose_constraint:  # and not np.isclose(dose, dose_constraint):
            # take larger steps using linear extrapolation
            delta_dose = self.get_dose(cfit, tau + 24.0, admin_datetime)[0] - dose
            slope = delta_dose / 24.0
            intercept = dose - (slope * tau)
            new_tau = (dose_constraint - intercept) / slope
            if (new_tau - tau) < 1.0:
                tau += 1.0
            else:
                tau = np.floor(new_tau)
            dose, tau = self.get_dose(cfit, tau, admin_datetime)

        while True:
            # walk it back
            if tau - 1.0 < 0.0:
                break
            if (
                self.get_dose(cfit, tau - 1.0, admin_datetime)[0] < dose_constraint
            ):  # or np.isclose(self.get_dose(cfit, tau - 1., admin_datetime)[0], dose_constraint):
                tau -= 1.0
                dose, tau = self.get_dose(cfit, tau, admin_datetime)
            else:
                break

        datetime_end = admin_datetime + timedelta(hours=tau)
        if next_element:
            for i in range(len(self.theta)):
                theta_hr = int(np.floor(self.theta[i]))
                theta_minute = int(np.floor((self.theta[i] - theta_hr) * 60.0))
                theta_datetime = datetime.combine(
                    datetime_end.date(), time(hour=theta_hr, minute=theta_minute)
                )
                if theta_datetime == datetime_end:
                    break
                elif theta_datetime > datetime_end:
                    extra_t = (theta_datetime - datetime_end).total_seconds() / 3600
                    tau += extra_t
                    datetime_end = theta_datetime
                    break
                elif i == len(self.theta) - 1:
                    theta_hr = int(np.floor(self.theta[0]))
                    theta_minute = int(np.floor((self.theta[0] - theta_hr) * 60.0))
                    theta_datetime = datetime.combine(
                        datetime_end.date(), time(hour=theta_hr, minute=theta_minute)
                    ) + timedelta(days=1)
                    extra_t = (theta_datetime - datetime_end).total_seconds() / 3600
                    tau += extra_t
                    datetime_end = theta_datetime

        return tau, dose, datetime_end

    def _get_restriction_arrays(
        self,
        cfit: clearance.Clearance_1m,
        dose_constraint,
        admin_datetime: datetime,
        next_element=True,
    ) -> tuple[float, float, np.ndarray, np.ndarray, datetime]:
        """Arrive at the restriction period (at most 1 h longer than it needs to be)
        by calculating the lifetime dose for delays in 1 h steps, starting from administration
        and stopping when the dose drops below the dose constraint
        (it might step once more to the start of the next pattern element if ``next_element`` True).

        * The delay is the time (h) from administration to when the contact pattern is resumed.
        * The restriction period is the least delay (up to the resolution of the pattern element widths) for which the lifetime dose received
          is less than the dose constraint.
        * Setting ``next_element`` True is advisable and can make a large difference for sparse repeating patterns such as public transport to and from work.
          E.g. If we want you to miss the morning bus, it means you have to wait for the afternoon bus;
          you can't just get on the morning bus at 9 AM instead of 8 AM and expect to comply with the dose constraint.
          By making the end of the restriction period coincide with the start of a pattern element,
          it is clear that contact can resume immediately at the end of the restriction period.
        * The calculated restriction period with ``next_element`` False is not necessarily equal to that calcuated by :meth:`get_restriction`
          with ``next_element`` False, though they have the same range.

        Args:
            cfit (clearance.Clearance_1m): Dose rate clearance object.
            dose_constraint (int or float): Dose constraint (mSv).
            admin_datetime (datetime.datetime): Administration datetime.
            next_element (bool, optional): If the end of the restriction period does not coincide with the start of a pattern element,
                extend the restriction period to the start of the next pattern element. Defaults to True.

        Raises:
            ValueError: If ``dose_constraint`` not greater than 0.

        Returns:
            tuple[float, float, numpy.ndarray, numpy.ndarray, datetime.datetime]: Calculated restriction period (h).
            With ``next_element`` False, this is in the range [true restriction period, true restriction period + (1 h) - min(``self``.c)[,
            where the true restriction period is limited only by the pattern element widths.

            Dose (mSv) from the pattern with the calculated restriction period;

            Delays (h) sampled up to the calculated restriction period;

            Doses (mSv) corresponding to the delays;

            Datetime at the end of the calculated restriction period.
        """
        if dose_constraint <= 0.0:
            raise ValueError("dose_constraint not greater than 0")

        tau_arr = []
        dose_arr = []

        tau = 0
        dose, tau = self.get_dose(cfit, tau, admin_datetime)
        tau_arr.append(tau)
        dose_arr.append(dose)

        while dose > dose_constraint:  # and not np.isclose(dose, dose_constraint):
            tau += 1.0
            dose, tau = self.get_dose(cfit, tau, admin_datetime)
            tau_arr.append(tau)
            dose_arr.append(dose)

        datetime_end = (
            None if admin_datetime == None else admin_datetime + timedelta(hours=tau)
        )
        if next_element:
            for i in range(len(self.theta)):
                theta_hr = int(np.floor(self.theta[i]))
                theta_minute = int(np.floor((self.theta[i] - theta_hr) * 60.0))
                theta_datetime = datetime.combine(
                    datetime_end.date(), time(hour=theta_hr, minute=theta_minute)
                )
                if theta_datetime == datetime_end:
                    break
                elif theta_datetime > datetime_end:
                    extra_t = (theta_datetime - datetime_end).total_seconds() / 3600
                    tau += extra_t
                    tau_arr.append(tau)
                    dose_arr.append(dose)
                    datetime_end = theta_datetime
                    break
                elif i == len(self.theta) - 1:
                    theta_hr = int(np.floor(self.theta[0]))
                    theta_minute = int(np.floor((self.theta[0] - theta_hr) * 60.0))
                    theta_datetime = datetime.combine(
                        datetime_end.date(), time(hour=theta_hr, minute=theta_minute)
                    ) + timedelta(days=1)
                    extra_t = (theta_datetime - datetime_end).total_seconds() / 3600
                    tau += extra_t
                    tau_arr.append(tau)
                    dose_arr.append(dose)
                    datetime_end = theta_datetime

        return tau, dose, np.array(tau_arr), np.array(dose_arr), datetime_end


class ContactPatternOnceoff(_ContactPattern):
    """Class for onceoff contact patterns."""

    def __init__(self, theta, c, d):
        """Constructor:
         * Calls constructor of :class:`_ContactPattern`.
         * ``theta``, ``c`` and ``d`` are converted into numpy.ndarrays if they were not supplied as such.

        Args:
            theta (int, float, list or numpy.ndarray): Time (h) **from end of delay** to start of pattern element.
            c (int, float, list or numpy.ndarray): Duration (h) of pattern element.
            d (int, float, list or numpy.ndarray): Distance (m) of pattern element;
              i.e., distance from radioactive person for duration of pattern element.

        Raises:
            ValueError: If ``theta`` not 0 for first element of onceoff pattern.
            ValueError: If element of onceoff pattern has ``d`` of 0.
        """
        super().__init__(theta, c, d)

        if isinstance(self.theta, (list, np.ndarray)):
            if self.theta[0] != 0:
                raise ValueError("theta not 0 for first element of onceoff pattern")
            if any(x == 0 for x in self.d):
                raise ValueError("element of onceoff pattern has d of 0")
        else:
            if self.theta != 0:
                raise ValueError(
                    "theta value not 0 for first element of onceoff pattern"
                )
            if self.d == 0:
                raise ValueError("element of onceoff pattern has d == 0")

        if isinstance(self.theta, list):
            self.theta = np.array(self.theta)
        elif isinstance(self.theta, (int, float)):
            self.theta = np.array([self.theta])
        if isinstance(self.c, list):
            self.c = np.array(self.c)
        elif isinstance(self.c, (int, float)):
            self.c = np.array([self.c])
        if isinstance(self.d, list):
            self.d = np.array(self.d)
        elif isinstance(self.d, (int, float)):
            self.d = np.array([self.d])

    def get_dose(self, cfit: clearance.Clearance_1m, tau, *args) -> float:
        """Return the dose (mSv) from sharing the onceoff pattern
        with a radioactive person, for a given dose rate clearance function and delay period.

        Unlike with repeating patterns, the dose from a onceoff pattern
        is independent of the administration datetime.
        ``*args`` allows this method to be called with additional argument `administration datetime`,
        so it can be called the same way on both :class:`ContactPatternRepeating` and :class:`ContactPatternOnceoff` objects.

        Args:
            cfit (clearance.Clearance_1m): Dose rate clearance from radioactive person.
            tau (int or float): Delay period (h); i.e., the time
                from administration to when the pattern is started.

        Raises:
            ValueError: If ``tau`` less than 0.

        Returns:
            float: Dose (mSv) from the pattern.
        """
        if tau < 0.0:
            raise ValueError("tau less than 0")

        if cfit.model == "exponential":
            dose_rate_1m_init, effective_half_life = cfit.model_params
            a = 1
            lmbda = np.log(2) / effective_half_life
            lmbda = np.array(
                [lmbda]
            )  # make it an ndarray of length n for n-component exponential
        elif cfit.model == "biexponential":
            dose_rate_1m_init, fraction_1, half_life_1, half_life_2 = cfit.model_params
            fraction_2 = 1.0 - fraction_1
            lmbda_1 = np.log(2) / half_life_1
            lmbda_2 = np.log(2) / half_life_2
            a = np.array([fraction_1, fraction_2])
            lmbda = np.array([lmbda_1, lmbda_2])

        arr = np.zeros(len(lmbda))
        for i in range(len(lmbda)):
            # summing over self.theta, self.c, self.d arrays (if more than 1 element in pattern, else np.sum does nothing)
            arr[i] = np.sum(
                self._dist_func(self.d)
                * np.exp(-lmbda[i] * self.theta)
                * (1 - np.exp(-lmbda[i] * self.c))
            )

        # summing over a, lmbda, arr arrays (if biexp, otherwise np.sum does nothing)
        dose = dose_rate_1m_init * np.sum((a / lmbda) * np.exp(-lmbda * tau) * arr)

        return dose / 1000.0  # uSv -> mSv

    def _exact_restriction_period(self, cfit: clearance.Clearance_1m, dose_constraint):
        """Return the exact restriction period required for a dose constraint.

        Args:
            cfit (clearance.Clearance_1m): Dose rate clearance object.
            dose_constraint (float): Dose constraint (mSv).

        Raises:
            ValueError: If ``dose_constraint`` not greater than 0.

        Returns:
            float: Exact restriction period (h).
        """
        if dose_constraint <= 0.0:
            raise ValueError("dose constraint not greater than 0")

        if self.get_dose(cfit, 0.0) < dose_constraint:
            warnings.warn(
                "no exact restriction period as dose is less than dose constraint with no delay"
            )
            return 0.0

        if cfit.model == "exponential":
            dose_rate_1m_init, effective_half_life = cfit.model_params
            lmbda = np.log(2) / effective_half_life
            rho = np.sum(
                self._dist_func(self.d)
                * np.exp(-lmbda * self.theta)
                * (1 - np.exp(-lmbda * self.c))
            )
            tau = (
                np.log(dose_rate_1m_init * rho / (dose_constraint * 1e3 * lmbda))
                / lmbda
            )  # uSv <-> mSv
            return tau
        elif cfit.model == "biexponential":
            return fsolve(lambda tau: self.get_dose(cfit, tau) - dose_constraint, 0.0)[
                0
            ]

    def get_restriction(
        self,
        cfit: clearance.Clearance_1m,
        dose_constraint,
        admin_datetime: datetime = None,
    ) -> tuple[float, float, datetime]:
        """Calculate the restriction period; i.e., the time from administration to beginning sharing the pattern with a
        radioactive person, such that the dose from the pattern is equal to the dose constraint.

        Args:
            cfit (clearance.Clearance_1m): Dose rate clearance from radioactive person.
            dose_constraint (int or float): Dose constraint (mSv).
            admin_datetime (datetime.datetime, optional): Administration datetime.

        Raises:
            ValueError: If ``dose_constraint`` not greater than 0.

        Returns:
            tuple[float, float, datetime.datetime]: Restriction period (h),
            dose (mSv) from the pattern with this restriction period, and
            datetime at the end of this restriction period (None if ``admin_datetime`` is None).
        """
        if dose_constraint <= 0.0:
            raise ValueError("dose_constraint not greater than 0")

        with warnings.catch_warnings():
            warnings.filterwarnings(
                "ignore",
                "no exact restriction period as dose is less than dose constraint with no delay",
            )
            tau = self._exact_restriction_period(cfit, dose_constraint)
        datetime_end = (
            None if admin_datetime == None else admin_datetime + timedelta(hours=tau)
        )
        return tau, self.get_dose(cfit, tau), datetime_end

    def _get_restriction_arrays(
        self,
        cfit: clearance.Clearance_1m,
        dose_constraint,
        admin_datetime: datetime = None,
    ) -> tuple[float, float, np.ndarray, np.ndarray, datetime]:
        """Arrive at the ceiling of the restriction period
        by calculating the dose for delays in 1 h steps, starting from administration
        and stopping when the dose drops below the dose constraint.

        * The delay is the time (h) from administration to when the contact pattern starts.
        * The restriction period is the delay for which the dose received
          is equal to the dose constraint.

        Args:
            cfit (clearance.Clearance_1m): Dose rate clearance object.
            dose_constraint (int or float): Dose constraint (mSv).
            admin_datetime (datetime.datetime, optional): Administration datetime.

        Raises:
            ValueError: If ``dose_constraint`` is not greater than 0.

        Returns:
            tuple[float, float, numpy.ndarray, numpy.ndarray, datetime.datetime]: Calculated restriction period (h),
            equal to the ceiling of the true restriction period;

            Dose (mSv) from the pattern with the calculated restriction period;

            Delays (h) sampled up to the calculated restriction period;

            Doses (mSv) corresponding to the delays;

            Datetime at the end of the calculated restriction period (None if ``admin_datetime`` is None).
        """
        if dose_constraint <= 0.0:
            raise ValueError("dose_constraint not greater than 0")

        tau_arr = []
        dose_arr = []

        tau = 0
        dose = self.get_dose(cfit, tau, admin_datetime)

        tau_arr.append(tau)
        dose_arr.append(dose)
        while dose > dose_constraint:  # and not np.isclose(dose, dose_constraint):
            tau += 1.0
            dose = self.get_dose(cfit, tau, admin_datetime)
            tau_arr.append(tau)
            dose_arr.append(dose)

        datetime_end = (
            None if admin_datetime == None else admin_datetime + timedelta(hours=tau)
        )
        return tau, dose, np.array(tau_arr), np.array(dose_arr), datetime_end


def cs_patterns() -> pd.DataFrame:
    """Return a dataframe containing the contact patterns published by Cormack & Shearer
    along with appropriate dose constraints.

    Reference: Cormack J & Shearer J. "Calculation of radiation exposures from patients to whom
    radioactive materials have been administered." Phys Med Biol 1998; 43(3).

    Returns:
        pandas.DataFrame:
        Dataframe with column labels:
         * `name` (str): A name for the contact pattern and dose constraint pairing.
         * `pattern_type` (str): 'repeating' or 'onceoff', indicating the type of pattern.
         * `theta` (int, float, list or numpy.ndarray): Start times (h) of pattern elements.
         * `c` (int, float, list or numpy.ndarray): Durations (h) of pattern elements.
         * `d` (int, float, list or numpy.ndarray): Distances (m) of pattern elements.
         * `dose_constraint` (int or float): Dose constraint (mSv).
         * `per_episode` (int): 1 if dose constraint is to be treated as per treatment episode, 0 if per annum.
    """
    list_of_dicts = []
    theta = np.linspace(0, 23, num=24)

    list_of_dicts.append(
        {
            "name": "Caring for infants (normal)",
            "pattern_type": "repeating",
            "theta": theta,
            "c": np.array(
                [
                    0,
                    0,
                    0,
                    20,
                    0,
                    0,
                    0,
                    20,
                    30,
                    40,
                    30,
                    20,
                    30,
                    40,
                    30,
                    20,
                    30,
                    40,
                    30,
                    20,
                    0,
                    0,
                    0,
                    35,
                ]
            )
            / 60.0,
            "d": [
                0,
                0,
                0,
                0.1,
                0,
                0,
                0,
                0.1,
                0.5,
                1,
                0.5,
                0.1,
                0.5,
                1,
                0.5,
                0.1,
                0.5,
                1,
                0.5,
                0.1,
                0,
                0,
                0,
                0.1,
            ],
            "dose_constraint": 1.0,
            "per_episode": 0,
        }
    )

    list_of_dicts.append(
        {
            "name": "Caring for infants (demanding or sick)",
            "pattern_type": "repeating",
            "theta": theta,
            "c": np.array(
                [
                    0,
                    0,
                    0,
                    35,
                    0,
                    0,
                    0,
                    35,
                    35,
                    35,
                    35,
                    35,
                    35,
                    35,
                    35,
                    35,
                    35,
                    35,
                    35,
                    35,
                    0,
                    0,
                    0,
                    35,
                ]
            )
            / 60.0,
            "d": [
                0,
                0,
                0,
                0.1,
                0,
                0,
                0,
                0.1,
                0.1,
                0.1,
                0.1,
                0.1,
                0.1,
                0.1,
                0.1,
                0.1,
                0.1,
                0.1,
                0.1,
                0.1,
                0,
                0,
                0,
                0.1,
            ],
            "dose_constraint": 1.0,
            "per_episode": 0,
        }
    )

    list_of_dicts.append(
        {
            "name": "Prolonged close contact (>15min) with 2-5 year old children",
            "pattern_type": "repeating",
            "theta": theta,
            "c": np.array(
                [
                    0,
                    0,
                    0,
                    0,
                    0,
                    0,
                    60,
                    30,
                    60,
                    60,
                    60,
                    60,
                    60,
                    60,
                    60,
                    60,
                    60,
                    30,
                    30,
                    60,
                    0,
                    0,
                    0,
                    0,
                ]
            )
            / 60.0,
            "d": [
                0,
                0,
                0,
                0,
                0,
                0,
                0.1,
                0.1,
                0.1,
                1,
                1,
                1,
                1,
                1,
                1,
                1,
                1,
                0.1,
                0.1,
                0.1,
                0,
                0,
                0,
                0,
            ],
            "dose_constraint": 1.0,
            "per_episode": 0,
        }
    )

    list_of_dicts.append(
        {
            "name": "Prolonged close contact (>15min) with 5-15 year old children",
            "pattern_type": "repeating",
            "theta": theta,
            "c": np.array(
                [
                    0,
                    0,
                    0,
                    0,
                    0,
                    0,
                    30,
                    30,
                    60,
                    0,
                    0,
                    0,
                    0,
                    0,
                    0,
                    0,
                    60,
                    60,
                    60,
                    30,
                    30,
                    0,
                    0,
                    0,
                ]
            )
            / 60.0,
            "d": [
                0,
                0,
                0,
                0,
                0,
                0,
                0.1,
                0.1,
                1,
                0,
                0,
                0,
                0,
                0,
                0,
                0,
                1,
                1,
                1,
                0.1,
                0.1,
                0,
                0,
                0,
            ],
            "dose_constraint": 1.0,
            "per_episode": 0,
        }
    )

    list_of_dicts.append(
        {
            "name": "Sleeping with another person (includes pregnant or child)",
            "pattern_type": "repeating",
            "theta": theta,
            "c": np.array(
                [
                    60,
                    60,
                    60,
                    60,
                    60,
                    60,
                    60,
                    60,
                    60,
                    0,
                    0,
                    0,
                    0,
                    0,
                    0,
                    0,
                    0,
                    0,
                    0,
                    60,
                    60,
                    60,
                    60,
                    60,
                ]
            )
            / 60.0,
            "d": [
                0.3,
                0.3,
                0.3,
                0.3,
                0.3,
                0.3,
                1,
                1,
                1,
                0,
                0,
                0,
                0,
                0,
                0,
                0,
                0,
                0,
                0,
                1,
                1,
                1,
                0.3,
                0.3,
            ],
            "dose_constraint": 1.0,
            "per_episode": 0,
        }
    )

    list_of_dicts.append(
        {
            "name": "Sleeping with informed supporter",
            "pattern_type": "repeating",
            "theta": theta,
            "c": np.array(
                [
                    60,
                    60,
                    60,
                    60,
                    60,
                    60,
                    60,
                    60,
                    60,
                    0,
                    0,
                    0,
                    0,
                    0,
                    0,
                    0,
                    0,
                    0,
                    0,
                    60,
                    60,
                    60,
                    60,
                    60,
                ]
            )
            / 60.0,
            "d": [
                0.3,
                0.3,
                0.3,
                0.3,
                0.3,
                0.3,
                1,
                1,
                1,
                0,
                0,
                0,
                0,
                0,
                0,
                0,
                0,
                0,
                0,
                1,
                1,
                1,
                0.3,
                0.3,
            ],
            "dose_constraint": 5.0,
            "per_episode": 1,
        }
    )

    list_of_dicts.append(
        {
            "name": "Sleeping with person and prolonged daytime close contact (>15min)",
            "pattern_type": "repeating",
            "theta": theta,
            "c": np.array(
                [
                    60,
                    60,
                    60,
                    60,
                    60,
                    60,
                    60,
                    60,
                    60,
                    60,
                    60,
                    60,
                    60,
                    60,
                    60,
                    60,
                    60,
                    60,
                    60,
                    60,
                    60,
                    60,
                    60,
                    60,
                ]
            )
            / 60.0,
            "d": [
                0.3,
                0.3,
                0.3,
                0.3,
                0.3,
                0.3,
                1,
                1,
                1,
                1,
                1,
                1,
                0.5,
                1,
                1,
                1,
                1,
                1,
                0.5,
                1,
                1,
                1,
                0.3,
                0.3,
            ],
            "dose_constraint": 1.0,
            "per_episode": 0,
        }
    )

    list_of_dicts.append(
        {
            "name": "Sleeping with informed supporter and prolonged daytime close contact (>15min)",
            "pattern_type": "repeating",
            "theta": theta,
            "c": np.array(
                [
                    60,
                    60,
                    60,
                    60,
                    60,
                    60,
                    60,
                    60,
                    60,
                    60,
                    60,
                    60,
                    60,
                    60,
                    60,
                    60,
                    60,
                    60,
                    60,
                    60,
                    60,
                    60,
                    60,
                    60,
                ]
            )
            / 60.0,
            "d": [
                0.3,
                0.3,
                0.3,
                0.3,
                0.3,
                0.3,
                1,
                1,
                1,
                1,
                1,
                1,
                0.5,
                1,
                1,
                1,
                1,
                1,
                0.5,
                1,
                1,
                1,
                0.3,
                0.3,
            ],
            "dose_constraint": 5.0,
            "per_episode": 1,
        }
    )

    list_of_dicts.append(
        {
            "name": "Prolonged close contact (>15min) with adult friends and family",
            "pattern_type": "repeating",
            "theta": theta,
            "c": np.array(
                [
                    0,
                    0,
                    0,
                    0,
                    0,
                    0,
                    0,
                    0,
                    30,
                    60,
                    60,
                    60,
                    60,
                    60,
                    60,
                    60,
                    60,
                    60,
                    60,
                    30,
                    60,
                    60,
                    0,
                    0,
                ]
            )
            / 60.0,
            "d": [
                0,
                0,
                0,
                0,
                0,
                0,
                0,
                0,
                0.5,
                1,
                1,
                1,
                0.5,
                1,
                1,
                1,
                1,
                1,
                0.5,
                0.1,
                1,
                1,
                0,
                0,
            ],
            "dose_constraint": 1.0,
            "per_episode": 0,
        }
    )

    list_of_dicts.append(
        {
            "name": "Prolonged close contact (>15min) with pregnant women",
            "pattern_type": "repeating",
            "theta": theta,
            "c": np.array(
                [
                    0,
                    0,
                    0,
                    0,
                    0,
                    0,
                    0,
                    0,
                    30,
                    60,
                    60,
                    60,
                    60,
                    60,
                    60,
                    60,
                    60,
                    60,
                    60,
                    30,
                    60,
                    60,
                    0,
                    0,
                ]
            )
            / 60.0,
            "d": [
                0,
                0,
                0,
                0,
                0,
                0,
                0,
                0,
                0.5,
                1,
                1,
                1,
                0.5,
                1,
                1,
                1,
                1,
                1,
                0.5,
                0.1,
                1,
                1,
                0,
                0,
            ],
            "dose_constraint": 1.0,
            "per_episode": 0,
        }
    )

    list_of_dicts.append(
        {
            "name": "Prolonged close contact (>15min) with informed persons caring for patient",
            "pattern_type": "repeating",
            "theta": theta,
            "c": np.array(
                [
                    0,
                    0,
                    0,
                    0,
                    0,
                    0,
                    30,
                    30,
                    60,
                    0,
                    0,
                    30,
                    30,
                    60,
                    0,
                    0,
                    60,
                    60,
                    60,
                    30,
                    30,
                    0,
                    0,
                    0,
                ]
            )
            / 60.0,
            "d": [
                0,
                0,
                0,
                0,
                0,
                0,
                0.1,
                0.5,
                1,
                0,
                0,
                0.1,
                0.5,
                1,
                0,
                0,
                1,
                1,
                1,
                0.5,
                0.1,
                0,
                0,
                0,
            ],
            "dose_constraint": 5.0,
            "per_episode": 1,
        }
    )

    list_of_dicts.append(
        {
            "name": "Cinema, theatre visits; social functions",
            "pattern_type": "repeating",
            "theta": 9.0,
            "c": 8.0,
            "d": 1.0,
            "dose_constraint": 1.0,
            "per_episode": 0,
        }
    )

    list_of_dicts.append(
        {
            "name": "Daily public transport to and from work",
            "pattern_type": "repeating",
            "theta": [8, 17],
            "c": [1, 1],
            "d": [0.3, 0.3],
            "dose_constraint": 1.0,
            "per_episode": 0,
        }
    )

    list_of_dicts.append(
        {
            "name": "Return to work involving prolonged close contact (>15min) with others",
            "pattern_type": "repeating",
            "theta": [9, 12, 14, 16],
            "c": [3, 2, 2, 1],
            "d": [0.5, 1, 0.5, 1],
            "dose_constraint": 1.0,
            "per_episode": 0,
        }
    )

    list_of_dicts.append(
        {
            "name": "Return to work not involving prolonged close contact (>15min) with others",
            "pattern_type": "repeating",
            "theta": 9.0,
            "c": 8.0,
            "d": 1.0,
            "dose_constraint": 1.0,
            "per_episode": 0,
        }
    )

    list_of_dicts.append(
        {
            "name": "Work with radiosensitive materials",
            "pattern_type": "repeating",
            "theta": 9.0,
            "c": 8.0,
            "d": 1.0,
            "dose_constraint": 0.1,
            "per_episode": 0,
        }
    )

    list_of_dicts.append(
        {
            "name": "Return to school",
            "pattern_type": "repeating",
            "theta": [9, 12, 14, 16],
            "c": [3, 2, 2, 1],
            "d": [0.5, 1, 0.5, 1],
            "dose_constraint": 1.0,
            "per_episode": 0,
        }
    )

    list_of_dicts.append(
        {
            "name": "A single 24-hour trip on public transport",
            "pattern_type": "onceoff",
            "theta": 0,
            "c": 24,
            "d": 0.3,
            "dose_constraint": 1.0,
            "per_episode": 0,
        }
    )

    df = pd.DataFrame.from_records(list_of_dicts)

    return df


def restrictions_for(
    df: pd.DataFrame,
    cfit: clearance.Clearance_1m,
    num_treatments_in_year,
    admin_datetime: datetime = None,
) -> pd.DataFrame:
    """Compute restriction periods for the supplied contact patterns and dose constraints in a dataframe,
    dose rate clearance function, number of treatments anticipated in a year and administration datetime.

    Args:
        df (pandas.DataFrame): Dataframe that includes column labels `pattern_type`, `theta`, `c`, `d`, `dose_constraint` and `per_episode`.
          See ``Returns`` section of :func:`cs_patterns`.
        cfit (clearance.Clearance_1m): Dose rate clearance from radioactive person.
        num_treatments_in_year (int, float): Number of treatments anticipated in a year.
          Used to scale `dose_constraint` if `per_episode` is 0.
        admin_datetime (datetime.datetime, optional): Administration datetime.
          Can only be omitted or None if `pattern_type` is 'onceoff' for all rows in ``df``,
          in which case the returned dataframe will have all None values in column `datetime_end`.

    Raises:
        KeyError: If ``df`` missing column label `pattern_type`, `theta`, `c`, `d`, `dose_constraint` or `per_episode`.
        ValueError: If row in ``df`` has `pattern_type` not 'repeating' or 'onceoff'.
        ValueError: If row in ``df`` has `pattern_type` 'repeating' and ``admin_datetime`` None.
        ValueError: If ``num_treatments_in_year`` less than 1.

    Returns:
        pandas.DataFrame: Deep copy of ``df`` with additional (or overwritten) columns labelled:
         * `dose_constraint_corrected` (float): Dose constraint (mSv) corrected for number of treatments in a year.
           Only differs from `dose_constraint` if `per_episode` is 0. If `per_episode` is 0, `dose_constraint` is per annum, so
           `dose_constraint_corrected` = `dose_constraint` / ``num_treatments_in_year``.
         * `restriction_period` (float): Calculated restriction period (h).
         * `dose` (float): Dose (mSv) from the pattern with the calculated restriction period.
         * `datetime_end` (datetime.datetime): Datetime at the end of the calculated restriction period
           (None if ``admin_datetime`` is None).
    """
    required_cols = [
        "pattern_type",
        "theta",
        "c",
        "d",
        "dose_constraint",
        "per_episode",
    ]
    if any(col not in df.columns for col in required_cols):
        raise KeyError("df missing column label in {}".format(required_cols))

    valid_pattern_types = ["repeating", "onceoff"]
    if any(p not in valid_pattern_types for p in df["pattern_type"].tolist()):
        raise ValueError(
            "row in df has pattern_type not in {}".format(valid_pattern_types)
        )

    if admin_datetime is None and "repeating" in df["pattern_type"].unique():
        raise ValueError("row in df has pattern_type repeating and admin_datetime None")

    if num_treatments_in_year < 1.0:
        raise ValueError("num_treatments_in_year less than 1")

    df = df.copy()

    df["dose_constraint_corrected"] = df["dose_constraint"] / num_treatments_in_year
    df.loc[df["per_episode"] == 1, ["dose_constraint_corrected"]] = df[
        "dose_constraint"
    ]

    tau_column = []
    dose_column = []
    datetime_end_column = []
    for _, row in df.iterrows():
        if row["pattern_type"] == "repeating":
            cpat = ContactPatternRepeating(row["theta"], row["c"], row["d"])
        elif row["pattern_type"] == "onceoff":
            cpat = ContactPatternOnceoff(row["theta"], row["c"], row["d"])

        tau, dose, datetime_end = cpat.get_restriction(
            cfit, row["dose_constraint_corrected"], admin_datetime
        )
        tau_column.append(tau)
        dose_column.append(dose)
        datetime_end_column.append(datetime_end)

    df["restriction_period"] = tau_column
    df["dose"] = dose_column
    df["datetime_end"] = datetime_end_column

    return df


def cs_restrictions(
    cfit: clearance.Clearance_1m, num_treatments_in_year, admin_datetime: datetime
) -> pd.DataFrame:
    """Return a dataframe containing the restriction periods for the contact patterns and dose constraints from Cormack & Shearer,
    using the supplied dose rate clearance function, number of treatments anticipated in a year and administration datetime.

    Reference: Cormack J & Shearer J. "Calculation of radiation exposures from patients to whom
    radioactive materials have been administered." Phys Med Biol 1998; 43(3).

    Args:
        cfit (clearance.Clearance_1m):  Dose rate clearance from radioactive person.
        num_treatments_in_year (int or float): Number of treatments anticipated in a year.
        admin_datetime (datetime.datetime): Administration datetime.

    Returns:
        pandas.DataFrame: Dataframe with column labels:
         * `name` (str): Name of the contact pattern and dose constraint pairing.
         * `pattern_type` (str): 'repeating' or 'onceoff', indicating the type of pattern.
         * `theta` (int, float, list or numpy.ndarray): Start times (h) of pattern elements.
         * `c` (int, float, list or numpy.ndarray): Durations (h) of pattern elements.
         * `d` (int, float, list or numpy.ndarray): Distances (m) of pattern elements.
         * `dose_constraint` (int or float): Dose constraint (mSv).
         * `per_episode` (int): 1 if dose constraint is to be treated as per treatment episode, 0 if per annum.
         * `dose_constraint_corrected` (float): Dose constraint (mSv) corrected for number of treatments in a year.
           Only differs from `dose_constraint` if `per_episode` is 0. If `per_episode` is 0, `dose_constraint` is per annum, so
           `dose_constraint_corrected` = `dose_constraint` / ``num_treatments_in_year``.
         * `restriction_period` (float): Calculated restriction period (h).
         * `dose` (float): Dose (mSv) from the pattern with the calculated restriction period.
         * `datetime_end` (datetime.datetime): Datetime at the end of the calculated restriction period.
    """
    return restrictions_for(
        cs_patterns(), cfit, num_treatments_in_year, admin_datetime=admin_datetime
    )



import importlib.metadata

try:
    __version__ = importlib.metadata.version("glowgreen")
except importlib.metadata.PackageNotFoundError:
    __version__ = None

from .clearance import Clearance_1m

from .close_contact import (
    ContactPatternRepeating,
    ContactPatternOnceoff,
    cs_restrictions,
    cs_patterns,
    restrictions_for,
)



from glowgreen import ContactPatternRepeating

theta = [7.5, 16]  # Start times (h) of pattern elements
c = [0.75, 3.5]  # Durations (h) of pattern elements
d = [0.3, 1]  # Distances (m) of pattern elements
cpat = ContactPatternRepeating(theta, c, d)

cpat.plot()

from glowgreen import Clearance_1m

dose_rate_init_2m = 60  # uSv/h
fraction1 = 0.3
half_life1 = 8  # h
half_life2 = 30  # h
distance = 2  # m
cfit = Clearance_1m(
    "biexponential", [dose_rate_init_2m, fraction1, half_life1, half_life2], distance
)

from datetime import datetime, timedelta

dose_constraint = 1  # mSv
admin_datetime = datetime(day=25, month=12, year=2021, hour=10, minute=30)
restriction_period, dose, datetime_end = cpat.get_restriction(
    cfit, dose_constraint, admin_datetime
)

assert dose <= dose_constraint
assert datetime_end == admin_datetime + timedelta(hours=restriction_period)

print(restriction_period, dose, datetime_end)

cpat.plot(cfit=cfit, dose_constraint=dose_constraint, admin_datetime=admin_datetime)

from glowgreen import cs_restrictions

num_treatments_in_year = 2
df = cs_restrictions(cfit, num_treatments_in_year, admin_datetime)

import pandas as pd
with pd.option_context('display.max_colwidth', None):
    print(df[["name", "datetime_end"]])



import pytest
from datetime import datetime
import numpy as np
from glowgreen.close_contact import _ContactPattern
from glowgreen import Clearance_1m, ContactPatternRepeating, ContactPatternOnceoff


def test_cpat_raises():
    with pytest.raises(TypeError):
        _ContactPattern([1, 2], [1, 2], 1)

    with pytest.raises(ValueError):
        _ContactPattern([1, 2], [1, 2], [1, 2, 3])

    with pytest.raises(TypeError):
        _ContactPattern([1, 2], [1, 2], ["1", 2])

    with pytest.raises(ValueError):
        _ContactPattern([1, 2], [-1, 2], [1, 2])

    with pytest.raises(ValueError):
        _ContactPattern([2, 1], [2, 1], [2, 1])

    with pytest.raises(ValueError):
        _ContactPattern([1, 2], [2, 2], [1, 2])

    with pytest.raises(TypeError):
        _ContactPattern(1, 1, [1, 2])

    with pytest.raises(ValueError):
        _ContactPattern(1, 1, -1)

    with pytest.raises(TypeError):
        _ContactPattern("1", 1, 1)


def test_plot():
    theta = np.array([0, 9.5])
    c = np.array([6.5, 2])
    d = [0.3, 1.5]
    cpat = ContactPatternOnceoff(theta, c, d)

    cpat.plot(test=True)
    cpat.plot(name="test", test=True)

    admin_datetime = datetime(year=2021, month=10, day=25, hour=10, minute=15)

    model = "exponential"
    dose_rate_xm_init = 60.0
    effective_half_life = 11.0
    model_parameters = [dose_rate_xm_init, effective_half_life]
    measurement_distance = 2.0
    cfit = Clearance_1m(model, model_parameters, measurement_distance)

    dose_constraint = 0.01

    cpat.plot(cfit=cfit, test=True)
    cpat.plot(cfit=cfit, dose_constraint=dose_constraint, test=True)

    cpat = ContactPatternRepeating(theta, c, d)
    cpat.plot(test=True)
    cpat.plot(cfit=cfit, dose_constraint=dose_constraint, test=True)
    cpat.plot(
        cfit=cfit,
        dose_constraint=dose_constraint,
        admin_datetime=admin_datetime,
        test=True,
    )



import pytest
import numpy as np
from datetime import datetime, timedelta
from glowgreen import ContactPatternRepeating, Clearance_1m


def test_rep_cpat_raises():
    with pytest.raises(ValueError):
        ContactPatternRepeating([0, 1], [1, 24], [1, 2])

    with pytest.raises(ValueError):
        ContactPatternRepeating(0, 25, 1)


def test_graduate_pattern():
    """Test ContactPatternRepeating.graduate_pattern, called in ContactPatternRepeating constructor."""
    theta = np.array([0.6, 2.5])
    c = np.array([0.7, 2])
    d = np.array([1, 2])
    cpat = ContactPatternRepeating(theta, c, d)
    assert np.array_equal(cpat.theta, np.array([0.6, 2.5, 3.5]))
    assert np.array_equal(cpat.c, np.array([0.7, 1.0, 1.0]))
    assert np.array_equal(cpat.d, np.array([1.0, 2.0, 2.0]))
    assert np.array_equal(theta, np.array([0.6, 2.5]))
    assert np.array_equal(c, np.array([0.7, 2]))
    assert np.array_equal(d, np.array([1, 2]))

    theta = [0.6, 2.5]
    c = [0.7, 2]
    d = [1, 2]
    cpat = ContactPatternRepeating(theta, c, d)
    assert np.array_equal(cpat.theta, np.array([0.6, 2.5, 3.5]))
    assert np.array_equal(cpat.c, np.array([0.7, 1.0, 1.0]))
    assert np.array_equal(cpat.d, np.array([1.0, 2.0, 2.0]))
    assert theta == [0.6, 2.5]
    assert c == [0.7, 2]
    assert d == [1, 2]

    theta = np.array([0.6, 1.4])
    c = np.array([0.7, 1.1])
    d = [1, 2]
    cpat = ContactPatternRepeating(theta, c, d)
    assert np.array_equal(cpat.theta, np.array([0.6, 1.4, 2.4]))
    assert np.allclose(cpat.c, np.array([0.7, 1.0, 0.1]))
    assert isinstance(cpat.c, np.ndarray)
    assert np.array_equal(cpat.d, np.array([1.0, 2.0, 2.0]))

    theta = [0.6, 1.4]
    c = [0.7, 1.1]
    d = np.array([1, 2])
    cpat = ContactPatternRepeating(theta, c, d)
    assert np.array_equal(cpat.theta, np.array([0.6, 1.4, 2.4]))
    assert np.allclose(cpat.c, np.array([0.7, 1.0, 0.1]))
    assert isinstance(cpat.c, np.ndarray)
    assert np.array_equal(cpat.d, np.array([1.0, 2.0, 2.0]))

    theta = [1.6, 2.9, 14.04]
    c = [0.3, 3.05, 1.31]
    d = [0.1, 1.9, 3.1]
    cpat = ContactPatternRepeating(theta, c, d)
    assert np.array_equal(cpat.theta, np.array([1.6, 2.9, 3.9, 4.9, 5.9, 14.04, 15.04]))
    assert np.allclose(cpat.c, np.array([0.3, 1.0, 1.0, 1.0, 0.05, 1, 0.31]))
    assert isinstance(cpat.c, np.ndarray)
    assert np.array_equal(cpat.d, np.array([0.1, 1.9, 1.9, 1.9, 1.9, 3.1, 3.1]))

    theta = 0.6
    c = 0.7
    d = 1
    cpat = ContactPatternRepeating(theta, c, d)
    assert np.array_equal(cpat.theta, np.array([0.6]))
    assert np.array_equal(cpat.c, np.array([0.7]))
    assert np.array_equal(cpat.d, np.array([1.0]))

    theta = [0.6]
    c = [0.7]
    d = [1]
    cpat = ContactPatternRepeating(theta, c, d)
    assert np.array_equal(cpat.theta, np.array([0.6]))
    assert np.array_equal(cpat.c, np.array([0.7]))
    assert np.array_equal(cpat.d, np.array([1.0]))

    theta = 1.3
    c = 0
    d = 0.4
    cpat = ContactPatternRepeating(theta, c, d)
    assert cpat.theta.size == 0
    assert cpat.c.size == 0
    assert cpat.d.size == 0

    theta = [1, 3, 4]
    c = [1.3, 0, 1]
    d = [0.4, 0, 0.6]
    cpat = ContactPatternRepeating(theta, c, d)
    assert np.array_equal(cpat.theta, np.array([1, 2, 4]))
    assert np.allclose(cpat.c, np.array([1, 0.3, 1]))
    assert np.array_equal(cpat.d, np.array([0.4, 0.4, 0.6]))

    theta = [1, 3, 4]
    c = [1.3, 0, 1]
    d = [0.4, 0, 0]
    with pytest.raises(ValueError):
        cpat = ContactPatternRepeating(theta, c, d)


def test_subtract_next_t_r_from():
    theta = np.array([5, 13])
    c = np.array([2, 3.5])
    d = [0.3, 1.5]
    cpat = ContactPatternRepeating(theta, c, d)

    datetime_admin = datetime(year=2021, month=10, day=25, hour=10, minute=15)

    assert cpat._subtract_next_t_r_from(datetime_admin) == -13.75


def test_get_dose_exp_1():
    dose_rate_init_1m = 50  # uSv/h
    effective_half_life = 40  # h
    tau = 0  # h

    # dose calculation by hand
    dose_hand = (
        dose_rate_init_1m * effective_half_life / (np.log(2) * 1e3)
    )  # uSv -> mSv

    # dose calculation by code
    theta = 0
    c = 24
    d = 1
    cpat = ContactPatternRepeating(theta, c, d)

    datetime_admin = datetime(year=2021, month=10, day=25, hour=10, minute=0)

    cfit = Clearance_1m("exponential", [dose_rate_init_1m, effective_half_life], 1.0)

    dose_code = cpat.get_dose(cfit, tau, datetime_admin)[0]

    assert np.isclose(dose_hand, dose_code)


def test_get_dose_exp_2():
    dose_rate_init_1m = 50  # uSv/h
    effective_half_life = 40  # h
    tau = 48  # h

    # dose calculation by hand
    dose_hand = (
        dose_rate_init_1m
        * effective_half_life
        * np.exp(-np.log(2) * tau / effective_half_life)
        / (np.log(2) * 1e3)
    )  # uSv -> mSv

    # dose calculation by code
    theta = 0
    c = 24
    d = 1
    cpat = ContactPatternRepeating(theta, c, d)

    datetime_admin = datetime(year=2021, month=10, day=25, hour=10, minute=0)

    cfit = Clearance_1m("exponential", [dose_rate_init_1m, effective_half_life], 1.0)

    dose_code = cpat.get_dose(cfit, tau, datetime_admin)[0]

    assert np.isclose(dose_hand, dose_code)


def test_get_dose_exp_3():
    """Test ContactPatternRepeating.get_dose method with exponential clearance function.

    Example problem:
      * Admin at 10:15 AM.
      * Dose rate at 1 m is (50 uSv/h)*exp[-ln2 t / (40 h)], where t is time from admin.
      * Contact pattern is repeating; 5 AM to 7 AM at 0.3 m, 1 PM to 4:30 PM at 1.5 m.
    Caclulate the dose for a restriction period of 51 h, so contact resumes at 1:15 PM 2 days later.

    Note the restriction period ends during a pattern element.
    The code breaks up the 13:00 to 16:30 PM pattern element into start times of 13, 14, 15 and 16.
    So the exposure from 1:15 to 2 PM of the first cycle will not be included.
    """
    dose_rate_init_1m = 50  # uSv/h
    effective_half_life = 40  # h
    tau = 51  # h

    # dose calculation by hand
    # phi = { (T_a - T_r) mod (p + tau)} mod p
    # phi = { (10.25 - 24) mod (24 + 51)} mod 24
    # phi = 13.25  # correct, restriction ends at 1:15 PM
    lmbda = np.log(2) / effective_half_life
    factor1 = (
        dose_rate_init_1m / lmbda * np.exp(-lmbda * tau) / (1 - np.exp(-lmbda * 24.0))
    )
    # 5 AM to 7 AM: theta_j - phi = 5 - 13.25 = -8.25, so S_p = -8.25 + 24 = 15.75, correct, this many hours from delay end to the first cycle (1:15 PM to 5 AM = 15.75 h)
    factor2 = 1 / (0.3) ** (1.5) * np.exp(-lmbda * 15.75) * (1 - np.exp(-lmbda * 2))
    # 1 PM to 4:30 PM: need to split up the contact, else the entire 3.5 hours of contact will be ignored for the first cycle, which is not what the code does
    # 1 PM to 2 PM:
    # theta_j - phi = 13 - 13.25 = -0.25, so S_p = -0.25 + 24 = 23.75, correct, this many hours from delay end to the first cycle (1:15 PM to 1 PM = 23.75 h)
    factor2 += 1 / (1.5) ** (1.5) * np.exp(-lmbda * 23.75) * (1 - np.exp(-lmbda * 1))
    # 2 PM to 4:30 PM: theta_j - phi = 14 - 13.25 = 0.75, so S_p = 0.75, correct, this many hours from delay end to the first cycle (1:15 PM to 2 PM = 0.75 h)
    factor2 += 1 / (1.5) ** (1.5) * np.exp(-lmbda * 0.75) * (1 - np.exp(-lmbda * 2.5))
    dose_hand = factor1 * factor2 / 1000.0  # uSv -> mSv

    # dose calculation by code
    theta = np.array([5, 13])
    c = np.array([2, 3.5])
    d = [0.3, 1.5]
    cpat = ContactPatternRepeating(theta, c, d)

    datetime_admin = datetime(year=2021, month=10, day=25, hour=10, minute=15)

    cfit = Clearance_1m("exponential", [dose_rate_init_1m, effective_half_life], 1.0)

    dose_code = cpat.get_dose(cfit, tau, datetime_admin)[0]
    assert np.isclose(dose_code, dose_hand)
    # ending tau at 2 PM gives the same dose
    assert np.isclose(cpat.get_dose(cfit, tau + 0.75, datetime_admin)[0], dose_code)
    # ending tau even slightly after 2PM means the exposure from 2 PM to 3 PM of the first cycle will not be included
    assert cpat.get_dose(cfit, tau + 0.76, datetime_admin)[0] < dose_code

    with pytest.raises(ValueError):
        cpat.get_dose(cfit, -0.1, datetime_admin)

    with pytest.raises(ValueError):
        cpat.get_dose(cfit, np.inf, datetime_admin)

    assert cpat.get_dose(cfit, 1e9, datetime_admin)[0] == 0.0

    with pytest.raises(ValueError):
        cpat.get_dose_finite(cfit, 2.0, 1.9, datetime_admin)
    assert np.isclose(
        cpat.get_dose_finite(cfit, tau, 1e9, datetime_admin)[0], dose_hand
    )
    assert (
        cpat.get_dose_finite(cfit, 0.0, 1e9, datetime_admin)[0]
        == cpat.get_dose(cfit, 0.0, datetime_admin)[0]
    )
    assert cpat.get_dose_finite(cfit, 1e9, 1e9, datetime_admin)[0] == 0.0


def test_get_dose_biexp_1():
    dose_rate_init_1m = 90  # uSv/h
    fraction1 = 0.3
    half_life1 = 5  # h
    half_life2 = 50  # h
    tau = 0  # h

    # dose calculation by hand
    dose_hand = (
        dose_rate_init_1m
        * (half_life1 * fraction1 + half_life2 * (1.0 - fraction1))
        / (np.log(2) * 1e3)
    )  # uSv -> mSv

    # dose calculation by code
    theta = 0
    c = 24
    d = 1
    cpat = ContactPatternRepeating(theta, c, d)

    datetime_admin = datetime(year=2021, month=10, day=25, hour=10, minute=0)

    cfit = Clearance_1m(
        "biexponential", [dose_rate_init_1m, fraction1, half_life1, half_life2], 1.0
    )

    dose_code = cpat.get_dose(cfit, tau, datetime_admin)[0]

    assert np.isclose(dose_hand, dose_code)


def test_get_dose_biexp_2():
    dose_rate_init_1m = 90  # uSv/h
    fraction1 = 0.3
    half_life1 = 5  # h
    half_life2 = 50  # h
    tau = 48  # h

    # dose calculation by hand
    dose_hand = (
        dose_rate_init_1m
        * (
            half_life1 * fraction1 * np.exp(-np.log(2) * tau / half_life1)
            + half_life2 * (1.0 - fraction1) * np.exp(-np.log(2) * tau / half_life2)
        )
        / (np.log(2) * 1e3)
    )  # uSv -> mSv

    # dose calculation by code
    theta = 0
    c = 24
    d = 1
    cpat = ContactPatternRepeating(theta, c, d)

    datetime_admin = datetime(year=2021, month=10, day=25, hour=10, minute=0)

    cfit = Clearance_1m(
        "biexponential", [dose_rate_init_1m, fraction1, half_life1, half_life2], 1.0
    )

    dose_code = cpat.get_dose(cfit, tau, datetime_admin)[0]

    assert np.isclose(dose_hand, dose_code)


def test_get_dose_biexp_3():
    """Test ContactPatternRepeating.get_dose method with biexponential clearance function.

    Example problem:
      * Admin at 10:15 AM.
      * Dose rate at 1 m is (90 uSv/h)*[0.3*exp(-ln2 t / (5 h)) + 0.7*exp(-ln2 t / (50 h))],
        where t is time from admin.
      * Contact pattern is repeating; 5 AM to 7 AM at 0.3 m, 1 PM to 4:30 PM at 1.5 m.
    Caclulate the dose for a restriction period of 51 h, so contact resumes at 1:15 PM 2 days later.

    Note the restriction period ends during a pattern element.
    The code breaks up the 13:00 to 16:30 PM pattern element into start times of 13, 14, 15 and 16.
    So the exposure from 1:15 to 2 PM of the first cycle will not be included.
    """
    dose_rate_init_1m = 90  # uSv/h
    fraction1 = 0.3
    half_life1 = 5  # h
    half_life2 = 50  # h
    tau = 51  # h

    # dose calculation by hand
    fraction2 = 1.0 - fraction1
    lmbda1 = np.log(2) / half_life1
    lmbda2 = np.log(2) / half_life2

    # first exp
    first_factor1 = (
        dose_rate_init_1m
        * fraction1
        / lmbda1
        * np.exp(-lmbda1 * tau)
        / (1 - np.exp(-lmbda1 * 24.0))
    )
    first_factor2 = (
        1 / (0.3) ** (1.5) * np.exp(-lmbda1 * 15.75) * (1 - np.exp(-lmbda1 * 2))
    )
    first_factor2 += (
        1 / (1.5) ** (1.5) * np.exp(-lmbda1 * 23.75) * (1 - np.exp(-lmbda1 * 1))
    )
    first_factor2 += (
        1 / (1.5) ** (1.5) * np.exp(-lmbda1 * 0.75) * (1 - np.exp(-lmbda1 * 2.5))
    )

    # second exp
    second_factor1 = (
        dose_rate_init_1m
        * fraction2
        / lmbda2
        * np.exp(-lmbda2 * tau)
        / (1 - np.exp(-lmbda2 * 24.0))
    )
    second_factor2 = (
        1 / (0.3) ** (1.5) * np.exp(-lmbda2 * 15.75) * (1 - np.exp(-lmbda2 * 2))
    )
    second_factor2 += (
        1 / (1.5) ** (1.5) * np.exp(-lmbda2 * 23.75) * (1 - np.exp(-lmbda2 * 1))
    )
    second_factor2 += (
        1 / (1.5) ** (1.5) * np.exp(-lmbda2 * 0.75) * (1 - np.exp(-lmbda2 * 2.5))
    )

    dose_hand = (
        (first_factor1 * first_factor2) + (second_factor1 * second_factor2)
    ) / 1000.0  # uSv -> mSv

    # dose calculation by code
    theta = np.array([5, 13])
    c = np.array([2, 3.5])
    d = [0.3, 1.5]
    cpat = ContactPatternRepeating(theta, c, d)

    datetime_admin = datetime(year=2021, month=10, day=25, hour=10, minute=15)

    cfit = Clearance_1m(
        "biexponential", [dose_rate_init_1m, fraction1, half_life1, half_life2], 1.0
    )

    dose_code = cpat.get_dose(cfit, tau, datetime_admin)[0]
    assert np.isclose(dose_code, dose_hand)
    # ending tau at 2 PM gives the same dose
    assert np.isclose(cpat.get_dose(cfit, tau + 0.75, datetime_admin)[0], dose_code)
    # ending tau even slightly after 2PM means the exposure from 2 PM to 3 PM of the first cycle will not be included
    assert cpat.get_dose(cfit, tau + 0.76, datetime_admin)[0] < dose_code

    with pytest.raises(ValueError):
        cpat.get_dose(cfit, -0.1, datetime_admin)

    with pytest.raises(ValueError):
        cpat.get_dose(cfit, np.inf, datetime_admin)[0]

    assert cpat.get_dose(cfit, 1e9, datetime_admin)[0] == 0.0

    with pytest.raises(ValueError):
        cpat.get_dose_finite(cfit, 2.0, 1.9, datetime_admin)
    assert np.isclose(
        cpat.get_dose_finite(cfit, tau, 1e9, datetime_admin)[0], dose_hand
    )
    assert (
        cpat.get_dose_finite(cfit, 0.0, 1e9, datetime_admin)[0]
        == cpat.get_dose(cfit, 0.0, datetime_admin)[0]
    )
    assert cpat.get_dose_finite(cfit, 1e9, 1e9, datetime_admin)[0] == 0.0


def test_get_restriction_exp():
    """Test ContactPatternRepeating.get_restriction and ContactPatternRepeating._get_restriction_arrays methods for
    exponential clearance function.

    For repeating patterns with next_element False,
    get_restriction and _get_restriction_arrays both return in the range:
    [true restriction period, true restriction period + (1 h) - min(self.c)[,
    where the true restriction period is limited only by the pattern element widths.
    But they do not necessarily return the same thing.

    Example problem:
      * Admin at 10:15 AM.
      * Dose rate at 1 m is (50 uSv/h)*exp[-ln2 t / (40 h)], where t is time from admin.
      * Contact pattern is repeating; 5 AM to 7 AM at 0.3 m, 1 PM to 4:30 PM at 1.5 m.
    Caclulate the dose for a restriction period of 51 h,
    so contact resumes at 1:15 PM 2 days later (= dose for delay of 51.75 h).
    Then use this dose (+ epsilon) as a dose constraint and calculate the restriction period.
    """
    dose_rate_init_1m = 50  # uSv/h
    effective_half_life = 40  # h
    tau_original = 51  # h

    # dose calculation by code
    theta = np.array([5, 13])
    c = np.array([2, 3.5])
    d = [0.3, 1.5]
    cpat = ContactPatternRepeating(theta, c, d)

    datetime_admin = datetime(year=2021, month=10, day=25, hour=10, minute=15)

    cfit = Clearance_1m("exponential", [dose_rate_init_1m, effective_half_life], 1.0)

    dose_code = cpat.get_dose(cfit, tau_original, datetime_admin)[0]
    dose_constraint = dose_code + 1e-9

    tau, dose, tau_arr, dose_arr, datetime_end = cpat._get_restriction_arrays(
        cfit, dose_constraint, datetime_admin, next_element=False
    )
    assert 51.75 <= tau < (51.75 + 1 - min(cpat.c))
    assert datetime_end == datetime_admin + timedelta(hours=tau)
    assert dose == cpat.get_dose(cfit, tau, datetime_admin)[0]
    assert dose <= dose_constraint
    assert tau_arr[-1] == tau
    assert dose_arr[-1] == dose
    assert len(tau_arr) == len(dose_arr)

    tau_fast, dose_fast, datetime_end_fast = cpat.get_restriction(
        cfit, dose_constraint, datetime_admin, next_element=False
    )
    assert 51.75 <= tau_fast < (51.75 + 1 - min(cpat.c))
    assert datetime_end_fast == datetime_admin + timedelta(hours=tau_fast)
    assert dose_fast == cpat.get_dose(cfit, tau_fast, datetime_admin)[0]
    assert dose_fast <= dose_constraint

    with pytest.raises(AttributeError):
        cpat._exact_restriction_period(cfit, dose_constraint)

    # For inf dose constraint,
    # In this case, admin_datetime is not during a pattern element, so
    # 	if next_element = False, restriction period will be 0 h
    # 	if next_element = True, restriction period will be 2.75 h (10:15 to 13:00)

    assert (
        cpat._get_restriction_arrays(cfit, np.inf, datetime_admin, next_element=False)[
            0
        ]
        == 0.0
    )
    assert (
        cpat._get_restriction_arrays(cfit, np.inf, datetime_admin, next_element=False)[
            1
        ]
        == cpat.get_dose(cfit, 0.0, datetime_admin)[0]
    )
    assert (
        cpat._get_restriction_arrays(cfit, np.inf, datetime_admin, next_element=False)[
            4
        ]
        == datetime_admin
    )
    assert (
        cpat._get_restriction_arrays(cfit, np.inf, datetime_admin, next_element=True)[0]
        == 2.75
    )
    assert np.isclose(
        cpat._get_restriction_arrays(cfit, np.inf, datetime_admin, next_element=True)[
            1
        ],
        cpat.get_dose(cfit, 2.75, datetime_admin)[0],
    )
    assert cpat._get_restriction_arrays(
        cfit, np.inf, datetime_admin, next_element=True
    )[4] == datetime_admin + timedelta(hours=2.75)

    # same for get_restriction
    assert (
        cpat.get_restriction(cfit, np.inf, datetime_admin, next_element=False)[0] == 0.0
    )
    assert (
        cpat.get_restriction(cfit, np.inf, datetime_admin, next_element=False)[1]
        == cpat.get_dose(cfit, 0.0, datetime_admin)[0]
    )
    assert (
        cpat.get_restriction(cfit, np.inf, datetime_admin, next_element=False)[2]
        == datetime_admin
    )
    assert (
        cpat.get_restriction(cfit, np.inf, datetime_admin, next_element=True)[0] == 2.75
    )
    assert np.isclose(
        cpat.get_restriction(cfit, np.inf, datetime_admin, next_element=True)[1],
        cpat.get_dose(cfit, 2.75, datetime_admin)[0],
    )
    assert cpat.get_restriction(cfit, np.inf, datetime_admin, next_element=True)[
        2
    ] == datetime_admin + timedelta(hours=2.75)

    with pytest.raises(ValueError):
        cpat._get_restriction_arrays(cfit, 0.0, datetime_admin)
    with pytest.raises(ValueError):
        cpat._get_restriction_arrays(cfit, -1.3, datetime_admin)
    with pytest.raises(ValueError):
        cpat.get_restriction(cfit, 0.0, datetime_admin)
    with pytest.raises(ValueError):
        cpat.get_restriction(cfit, -1.3, datetime_admin)


def test_get_restriction_biexp():
    """Test get_restriction and _get_restriction_arrays methods for
    biexponential clearance function.

    For repeating patterns with next_element False,
    get_restriction and _get_restriction_arrays both return in the range:
    [true restriction period, true restriction period + (1 h) - min(self.c)[,
    where the true restriction period is limited only by the pattern element widths.
    But they do not necessarily return the same thing.
    Further, they do not necessarily return the same dose,
    though each dose will be less than the dose constraint.

    Example problem:
      * Admin at 10:15 AM.
      * Dose rate at 1 m is (90 uSv/h)*[0.3*exp(-ln2 t / (5 h)) + 0.7*exp(-ln2 t / (50 h))],
        where t is time from admin.
      * Contact pattern is repeating; 5 AM to 7 AM at 0.3 m, 10 AM to 11 AM at 1 m, 4:20 PM to 4:30 PM at 0.1 m.
    Caclulate the dose for a restriction period of 54.25 h,
    so contact resuming at 4:30 PM 2 days later.
    Then use this dose (+ epsilon) as a dose constraint and calculate the restriction period.
    """
    dose_rate_init_1m = 90  # uSv/h
    fraction1 = 0.3
    half_life1 = 5  # h
    half_life2 = 50  # h

    tau_original = 54.25  # h

    # dose calculation by code
    theta = np.array([5, 10, 16 + 2 / 6])
    c = np.array([2, 1, 1 / 6])
    d = [0.3, 1, 0.1]
    cpat = ContactPatternRepeating(theta, c, d)

    datetime_admin = datetime(year=2021, month=10, day=25, hour=10, minute=15)

    cfit = Clearance_1m(
        "biexponential", [dose_rate_init_1m, fraction1, half_life1, half_life2], 1.0
    )

    dose_code = cpat.get_dose(cfit, tau_original, datetime_admin)[0]
    dose_constraint = dose_code + 1e-9

    tau, dose, tau_arr, dose_arr, datetime_end = cpat._get_restriction_arrays(
        cfit, dose_constraint, datetime_admin, next_element=False
    )
    assert 54.25 <= tau <= (54.25 + 1 - min(cpat.c))
    assert datetime_end == datetime_admin + timedelta(hours=tau)
    assert dose == cpat.get_dose(cfit, tau, datetime_admin)[0]
    assert dose <= dose_constraint
    assert tau_arr[-1] == tau
    assert dose_arr[-1] == dose
    assert len(tau_arr) == len(dose_arr)

    tau_fast, dose_fast, datetime_end_fast = cpat.get_restriction(
        cfit, dose_constraint, datetime_admin, next_element=False
    )
    assert 54.25 <= tau_fast <= (54.25 + 1 - min(cpat.c))
    assert datetime_end_fast == datetime_admin + timedelta(hours=tau_fast)
    assert np.isclose(dose_fast, cpat.get_dose(cfit, tau_fast, datetime_admin)[0])
    assert dose_fast <= dose_constraint

    # next_element True
    tau_next, dose_next, _, _, datetime_end_next = cpat._get_restriction_arrays(
        cfit, dose_constraint, datetime_admin, next_element=True
    )
    assert tau_next == 66.75
    assert dose_next == cpat.get_dose(cfit, 66.75, datetime_admin)[0]
    assert datetime_end_next == datetime_admin + timedelta(hours=66.75)

    tau_fast_next, dose_fast_next, datetime_end_fast_next = cpat.get_restriction(
        cfit, dose_constraint, datetime_admin, next_element=True
    )
    assert tau_fast_next == 66.75
    assert np.isclose(dose_fast_next, cpat.get_dose(cfit, 66.75, datetime_admin)[0])
    assert datetime_end_fast_next == datetime_admin + timedelta(hours=66.75)

    # inf dose constraint
    # In this case, admin_datetime occurs during a contact element, so
    # 	if next_element = False, restriction period will be 0.75 h.
    # 	If next_element = True, restriction period will be 0.75 + 5 + 19/60 h (10:15 AM to 16:19)
    # 	or 0.75 + 5 + 1/3 h (10:15 AM to 16:20), depending how a float appears
    assert (
        cpat._get_restriction_arrays(cfit, np.inf, datetime_admin, next_element=False)[
            0
        ]
        == 0.75
    )
    assert (
        cpat._get_restriction_arrays(cfit, np.inf, datetime_admin, next_element=False)[
            1
        ]
        == cpat.get_dose(cfit, 0.75, datetime_admin)[0]
    )
    assert cpat._get_restriction_arrays(
        cfit, np.inf, datetime_admin, next_element=False
    )[4] == datetime_admin + timedelta(hours=0.75)
    assert cpat._get_restriction_arrays(
        cfit, np.inf, datetime_admin, next_element=True
    )[0] in [0.75 + 5 + 19 / 60, 0.75 + 5 + 1 / 3]
    assert np.isclose(
        cpat._get_restriction_arrays(cfit, np.inf, datetime_admin, next_element=True)[
            1
        ],
        cpat.get_dose(cfit, 0.75 + 5 + 1 / 3, datetime_admin)[0],
    )
    assert cpat._get_restriction_arrays(
        cfit, np.inf, datetime_admin, next_element=True
    )[4] in [
        datetime_admin + timedelta(hours=0.75 + 5 + 19 / 60),
        datetime_admin + timedelta(hours=0.75 + 5 + 1 / 3),
    ]

    # same for get_restriction
    assert (
        cpat.get_restriction(cfit, np.inf, datetime_admin, next_element=False)[0]
        == 0.75
    )
    assert (
        cpat.get_restriction(cfit, np.inf, datetime_admin, next_element=False)[1]
        == cpat.get_dose(cfit, 0.75, datetime_admin)[0]
    )
    assert cpat.get_restriction(cfit, np.inf, datetime_admin, next_element=False)[
        2
    ] == datetime_admin + timedelta(hours=0.75)
    assert cpat.get_restriction(cfit, np.inf, datetime_admin, next_element=True)[0] in [
        0.75 + 5 + 19 / 60,
        0.75 + 5 + 1 / 3,
    ]
    assert np.isclose(
        cpat.get_restriction(cfit, np.inf, datetime_admin, next_element=True)[1],
        cpat.get_dose(cfit, 0.75 + 5 + 1 / 3, datetime_admin)[0],
    )
    assert cpat.get_restriction(cfit, np.inf, datetime_admin, next_element=True)[2] in [
        datetime_admin + timedelta(hours=0.75 + 5 + 19 / 60),
        datetime_admin + timedelta(hours=0.75 + 5 + 1 / 3),
    ]
